from xagents import a2c

a2c_args = a2c.cli_args
ppo_args = {
    'model': {'help': 'Path to model .cfg file'},
    'lam': {
        'help': 'GAE-Lambda for advantage estimation',
        'type': float,
        'default': 0.95,
        'hp_type': 'log_uniform',
    },
    'ppo-epochs': {
        'help': 'Gradient updates per training step',
        'type': int,
        'default': 4,
        'hp_type': 'categorical',
    },
    'mini-batches': {
        'help': 'Number of mini-batches to use per update',
        'type': int,
        'default': 4,
        'hp_type': 'categorical',
    },
    'advantage-epsilon': {
        'help': 'Value added to estimated advantage',
        'type': float,
        'default': 1e-8,
        'hp_type': 'log_uniform',
    },
    'clip-norm': {
        'help': 'Clipping value passed to tf.clip_by_value()',
        'type': float,
        'default': 0.1,
        'hp_type': 'log_uniform',
    },
    'n-steps': {
        'help': 'Transition steps',
        'type': int,
        'default': 128,
        'hp_type': 'categorical',
    },
}
cli_args = a2c_args.copy()
cli_args.update(ppo_args)



cli_args = {
    'model': {'help': 'Path to model .cfg file'},
    'double': {'help': 'If specified, DDQN will be used', 'action': 'store_true'},
    'epsilon-start': {
        'help': 'Starting epsilon value which is used to control random exploration.\n'
        'It should be decremented and adjusted according to implementation needs',
        'type': float,
        'default': 1.0,
        'hp_type': 'log_uniform',
    },
    'epsilon-end': {
        'help': 'Epsilon end value (minimum exploration rate)',
        'type': float,
        'default': 0.02,
        'hp_type': 'log_uniform',
    },
    'epsilon-decay-steps': {
        'help': 'Number of steps for `epsilon-start` to reach `epsilon-end`',
        'type': float,
        'default': 150000,
        'hp_type': 'int',
    },
    'target-sync-steps': {
        'help': 'Sync target models every n steps',
        'type': int,
        'default': 1000,
        'hp_type': 'int',
    },
}



import argparse
import sys
import warnings

import pandas as pd

import xagents
from xagents.base import OffPolicy
from xagents.utils.cli import agent_args, non_agent_args, off_policy_args
from xagents.utils.common import create_agent
from xagents.utils.tuning import run_tuning


class Executor:
    """
    Command line parser.
    """

    def __init__(self):
        """
        Initialize supported commands and agents.
        """
        self.agent_id = None
        self.command = None
        self.agent = None

    @staticmethod
    def display_section(title, cli_args):
        """
        Display given title (command) and respective available options.
        Args:
            title: Command(s) that will be displayed on top of cli options.
            cli_args: A dictionary having flags and their respective
                `help`, `required` and `default`

        Returns:
            None
        """
        section_frame = pd.DataFrame(cli_args).T.fillna('-')
        section_frame['flags'] = section_frame.index.values
        section_frame['flags'] = section_frame['flags'].apply(lambda flag: f'--{flag}')
        section_frame = (
            section_frame.reset_index(drop=True).set_index('flags').sort_index()
        )
        print(f'\n{title}\n')
        print(
            section_frame[
                [
                    column_name
                    for column_name in ('help', 'default', 'hp_type')
                    if column_name in section_frame.columns
                ]
            ].to_markdown()
        )

    def display_commands(self, sections=None):
        """
        Display available commands and their description
            + command specific sections if given any.
        Args:
            sections: A dictionary having flags and their respective
                `help`, `required` and `default`

        Returns:
            None
        """
        print(f'xagents {xagents.__version__}')
        print(f'\nUsage:')
        print(f'\txagents <command> <agent> [options] [args]')
        print(f'\nAvailable commands:')
        for command, items in xagents.commands.items():
            print(f'\t{command:<10} {items[2]}')
        print()
        print('Use xagents <command> to see more info about a command')
        print('Use xagents <command> <agent> to see more info about command + agent')
        if sections:
            for title, cli_args in sections.items():
                self.display_section(title, cli_args)

    @staticmethod
    def add_args(cli_args, parser, tuning=False):
        """
        Add given arguments to parser.
        Args:
            cli_args: A dictionary of args and options.
            parser: argparse.ArgumentParser
            tuning: If True, flags that have an `hp_type`
                will support multiple parameters which will
                be parsed according to their type.

        Returns:
            None.
        """
        for arg, options in cli_args.items():
            _help = options.get('help')
            _default = options.get('default')
            _type = options.get('type')
            _action = options.get('action')
            _required = options.get('required')
            _nargs = options.get('nargs')
            _hp_type = options.get('hp_type')
            if not _action:
                if tuning and _hp_type:
                    _nargs = '*'
                parser.add_argument(
                    f'--{arg}',
                    help=_help,
                    default=_default,
                    type=_type,
                    required=_required,
                    nargs=_nargs,
                )
            else:
                parser.add_argument(
                    f'--{arg}', help=_help, default=_default, action=_action
                )

    def maybe_create_agent(self, argv):
        """
        Display help respective to parsed commands or set self.agent_id and self.command
        for further execution if enough arguments are given.
        Args:
            argv: Arguments passed through sys.argv or otherwise.

        Returns:
            None
        """
        to_display = {}
        total = len(argv)
        if total == 0:
            self.display_commands()
            return
        command = argv[0]
        to_display.update(non_agent_args)
        to_display.update(agent_args)
        assert command in xagents.commands, f'Invalid command `{command}`'
        to_display.update(xagents.commands[command][0])
        if total == 1:
            self.display_commands({command: to_display})
            return
        agent_id = argv[1]
        assert agent_id in xagents.agents, f'Invalid agent `{agent_id}`'
        to_display.update(xagents.agents[agent_id]['module'].cli_args)
        if total == 2:
            title = f'{command} {agent_id}'
            if (
                issubclass(xagents.agents[agent_id]['agent'], OffPolicy)
                or agent_id == 'acer'
            ):
                to_display.update(off_policy_args)
            self.display_commands({title: to_display})
            return
        self.command, self.agent_id = command, agent_id

    def parse_known_args(self, argv, tuning=False):
        """
        Parse general, agent and command specific args.
        Args:
            argv: Arguments passed through sys.argv or otherwise.
            tuning: If True, flags that have an `hp_type`
                will support multiple parameters which will
                be parsed according to their type.
        Returns:
            agent kwargs, non-agent kwargs and command kwargs.
        """
        general_parser = argparse.ArgumentParser()
        agent_parser = argparse.ArgumentParser()
        command_parser = argparse.ArgumentParser()
        self.add_args(agent_args, agent_parser, tuning)
        self.add_args(
            xagents.agents[self.agent_id]['module'].cli_args, agent_parser, tuning
        )
        self.add_args(xagents.commands[self.command][0], command_parser, tuning)
        if (
            issubclass(xagents.agents[self.agent_id]['agent'], OffPolicy)
            or self.agent_id == 'acer'
        ):
            self.add_args(off_policy_args, general_parser, tuning)
        self.add_args(non_agent_args, general_parser, tuning)
        non_agent_known, extra1 = general_parser.parse_known_args(argv)
        agent_known, extra2 = agent_parser.parse_known_args(argv)
        command_known, extra3 = command_parser.parse_known_args(argv)
        unknown_flags = [
            unknown_flag
            for unknown_flag in set(extra1) & set(extra2) & set(extra3)
            if unknown_flag not in [self.command, self.agent_id]
            and '--' in unknown_flag
        ]
        if unknown_flags:
            warnings.warn(f'Got unknown flags {unknown_flags}')
        if self.command == 'train':
            assert (
                command_known.target_reward or command_known.max_steps
            ), 'train requires --target-reward or --max-steps'
        return agent_known, non_agent_known, command_known

    def execute(self, argv):
        """
        Parse command line arguments, display help or execute command
        if enough arguments are given.
        Args:
            argv: Arguments passed through sys.argv or otherwise.

        Returns:
            None
        """
        self.maybe_create_agent(argv)
        if not self.agent_id:
            return
        if self.command == 'tune':
            agent_known, non_agent_known, command_known = self.parse_known_args(
                argv, True
            )
            run_tuning(self.agent_id, agent_known, non_agent_known, command_known)
        else:
            agent_known, non_agent_known, command_known = self.parse_known_args(
                argv,
            )
            self.agent = create_agent(
                self.agent_id, vars(agent_known), vars(non_agent_known)
            )
            getattr(self.agent, xagents.commands[self.command][1])(
                **vars(command_known)
            )


def execute(argv=None):
    """
    Parse and execute commands.
    Args:
        argv: List of arguments to be passed to Executor.execute()
            if not specified, defaults. to sys.argv[1:]

    Returns:
        None
    """
    argv = argv or sys.argv[1:]
    Executor().execute(argv)


if __name__ == '__main__':
    execute()



cli_args = {
    'actor-model': {'help': 'Path to actor model .cfg file'},
    'critic-model': {'help': 'Path to critic model .cfg file'},
    'gradient-steps': {
        'help': 'Number of iterations per train step',
        'type': int,
        'hp_type': 'int',
    },
    'tau': {
        'help': 'Value used for syncing target model weights',
        'type': float,
        'default': 0.005,
        'hp_type': 'log_uniform',
    },
    'step-noise-coef': {
        'help': 'Coefficient multiplied by noise added to actions to step',
        'type': float,
        'default': 0.1,
        'hp_type': 'log_uniform',
    },
}



import numpy as np
import tensorflow as tf
from gym.spaces import Discrete

from xagents import A2C
from xagents.utils.common import LazyFrames


class ACER(A2C):
    """
    Sample Efficient Actor-Critic with Experience Replay.
    https://arxiv.org/abs/1611.01224
    """

    def __init__(
        self,
        envs,
        model,
        buffers,
        ema_alpha=0.99,
        replay_ratio=4,
        epsilon=1e-6,
        importance_c=10.0,
        delta=1,
        trust_region=True,
        **kwargs,
    ):
        """
        Initialize ACER agent.
        Args:
            envs: A list of gym environments.
            model: tf.keras.models.Model that is expected to be compiled
                with an optimizer before training starts.
            buffers: A list of replay buffer objects whose length should match
                `envs`s'.
            ema_alpha: Moving average decay passed to tf.train.ExponentialMovingAverage()
            replay_ratio: Lam value passed to np.random.poisson()
            epsilon: epsilon value used in several calculations during gradient update.
            importance_c: Importance weight truncation parameter.
            delta: Delta parameter used for trust region update.
            trust_region: If False, no trust region updates will be used.
            **kwargs: kwargs Passed to super classes.
        """
        super(ACER, self).__init__(envs, model, **kwargs)
        self.assert_valid_env(envs[0], Discrete)
        self.avg_model = tf.keras.models.clone_model(self.model)
        self.ema = tf.train.ExponentialMovingAverage(ema_alpha)
        self.buffers = buffers
        assert (
            buffers[0].batch_size == 1
        ), f'Buffer batch size should be 1 for ACER, got {buffers[0].batch_size}'
        self.buffer_current_size = tf.Variable(0)
        self.batch_indices = tf.range(self.n_steps * self.n_envs, dtype=tf.int64)[
            :, tf.newaxis
        ]
        self.replay_ratio = replay_ratio
        self.epsilon = epsilon
        self.importance_c = importance_c
        self.delta = delta
        self.trust_region = trust_region
        self.batch_dtypes = ['uint8', 'float32', 'int32', 'float32', 'float32']
        self.batch_shapes = [
            (self.n_envs * (self.n_steps + 1), *self.input_shape),
            (self.n_envs * self.n_steps),
            (self.n_envs * self.n_steps),
            (self.n_envs * self.n_steps),
            (self.n_envs * self.n_steps, self.n_actions),
        ]

    def flat_to_steps(self, t, steps=None):
        """
        Split a tensor to tensors of shape (self.n_envs, self.n_steps)
        Args:
            t: A flat tensor that has (self.n_envs * self.n_steps) values.
            steps: If not specified, self.n_steps is used by default.

        Returns:
            A list of self.n_envs x self.n_steps tensors.
        """
        t = tf.reshape(t, (self.n_envs, steps or self.n_steps, *t.shape[1:]))
        return [
            tf.squeeze(step_t, 1) for step_t in tf.split(t, steps or self.n_steps, 1)
        ]

    def clip_last_step(self, t):
        """
        Remove last step from a given tensor.
        Args:
            t: Tensor that has self.n_steps + 1 items.

        Returns:
            Tensor that has self.n_steps items.
        """
        ts = self.flat_to_steps(t, self.n_steps + 1)
        return tf.reshape(tf.stack(ts[:-1], 1), (-1, *ts[0].shape[1:]))

    @staticmethod
    def add_grads(g1, g2):
        """
        Add 2 gradients if neither is None, otherwise return one of them.
        Args:
            g1: Grad 1
            g2: Grad 2

        Returns:
            One of them or their sum.
        """
        assert g1 is not None or g2 is not None, 'Gradients are None'
        if g1 is not None and g2 is not None:
            return g1 + g2
        if g1 is not None:
            return g1
        return g2

    def update_avg_weights(self):
        """
        Update average model weights after performing gradient update.

        Returns:
            None
        """
        avg_variables = [
            self.ema.average(weight).numpy()
            for weight in self.model.trainable_variables
        ]
        self.avg_model.set_weights(avg_variables)

    def store_batch(self, batch):
        """
        Store batch in self.buffers.
        Args:
            batch: A list of (states, rewards, actions, dones, actor output)

        Returns:
            None
        """
        for i in range(self.n_envs):
            env_outputs = []
            for item in batch:
                current = item[i]
                if len(current.shape) > 2:
                    current = LazyFrames(current.astype(np.uint8))
                env_outputs.append(current)
            self.buffers[i].append(*env_outputs)

    def get_batch(self):
        """
        Get a batch of (states, rewards, actions, dones, actor output),
        save batch to replay buffers and adjust shapes for gradient update.

        Returns:
            Merged environment outputs.
        """
        (
            states,
            rewards,
            actions,
            _,
            dones,
            *_,
            actor_logits,
        ) = super(ACER, self).get_batch()
        states.append(self.get_states())
        batch = [states, rewards, actions, dones[1:], actor_logits]
        batch = [np.asarray(item).swapaxes(0, 1) for item in batch]
        self.store_batch(batch)
        return [
            item.reshape(-1, *item.shape[2:]).astype(dtype)
            for (item, dtype) in zip(batch, self.batch_dtypes)
        ]

    def calculate_returns(
        self,
        rewards,
        dones,
        values=None,
        selected_critic_logits=None,
        selected_importance=None,
    ):
        """
        Get a batch of returns.
        Args:
            rewards: Rewards tensor of shape (self.n_steps, self.n_envs)
            dones: Dones tensor of shape (self.n_steps, self.n_envs)
            values: Values tensor of shape (self.n_steps + 1, self.n_envs).
                required for PPO, TRPO and ACER
            selected_critic_logits: Critic output respective to selected actions
                of shape (self.n_steps, self.n_envs).
                Required for ACER.
            selected_importance: Importance weights respective to selected
                actions of shape (self.n_steps, self.n_envs).
                Required for ACER
        Returns:
            Tensor of n-step returns.
        """
        importance_bar = self.flat_to_steps(tf.minimum(1.0, selected_importance))
        dones = self.flat_to_steps(dones)
        rewards = self.flat_to_steps(rewards)
        selected_critic_logits = self.flat_to_steps(selected_critic_logits)
        values = self.flat_to_steps(values, self.n_steps + 1)
        current_return = values[-1]
        returns = []
        for i in reversed(range(self.n_steps)):
            current_return = rewards[i] + self.gamma * current_return * (1.0 - dones[i])
            returns.append(current_return)
            current_return = (
                importance_bar[i] * (current_return - selected_critic_logits[i])
            ) + values[i]
        return tf.reshape(tf.stack(returns[::-1], 1), [-1])

    def calculate_losses(
        self,
        action_probs,
        values,
        returns,
        selected_probs,
        selected_importance,
        selected_critic_logits,
    ):
        """
        Calculate model loss.
        Args:
            action_probs: A tensor of shape (self.n_envs * self.n_steps, self.n_actions)
            values: A tensor of shape self.n_envs * (self.n_steps + 1)
            returns: A tensor of shape (self.n_envs * self.n_steps)
            selected_probs: A tensor of shape (self.n_envs * self.n_steps)
            selected_importance: A tensor of shape (self.n_envs * self.n_steps)
            selected_critic_logits: A tensor of shape (self.n_envs * self.n_steps)

        Returns:
            Loss as a TF tensor if trust region is not used, otherwise loss, value_loss.
        """
        entropy = tf.reduce_mean(
            -tf.reduce_sum(
                action_probs * tf.math.log(action_probs + self.epsilon), axis=1
            )
        )
        values = self.clip_last_step(values)
        advantages = returns - values
        log_probs = tf.math.log(selected_probs + self.epsilon)
        action_gain = log_probs * tf.stop_gradient(
            advantages * tf.minimum(self.importance_c, selected_importance)
        )
        action_loss = -tf.reduce_mean(action_gain)
        value_loss = (
            tf.reduce_mean(
                tf.square(tf.stop_gradient(returns) - selected_critic_logits) * 0.5
            )
            * self.value_loss_coef
        )
        if self.trust_region:
            return (
                -(action_loss - self.entropy_coef * entropy)
                * self.n_steps
                * self.n_envs
            ), value_loss
        return (
            action_loss
            + self.value_loss_coef * value_loss
            - self.entropy_coef * entropy
        )

    def calculate_grads(self, tape, losses, action_probs, avg_action_probs):
        """
        Calculate gradients given loss(es)
        Args:
            tape: tf.GradientTape()
            losses: loss or loss, value_loss (if trust region is used)
            action_probs: A tensor of shape (self.n_envs * self.n_steps, self.n_actions)
            avg_action_probs:  A tensor of shape (self.n_envs * self.n_steps, self.n_actions)

        Returns:
            A list of gradients.
        """
        if not self.trust_region:
            return tape.gradient(losses, self.model.trainable_variables)
        loss, value_loss = losses
        g = tape.gradient(
            loss,
            action_probs,
        )
        k = -avg_action_probs / (action_probs + self.epsilon)
        adj = tf.maximum(
            0.0,
            (tf.reduce_sum(k * g, axis=-1) - self.delta)
            / (tf.reduce_sum(tf.square(k), axis=-1) + self.epsilon),
        )
        g = g - tf.reshape(adj, [self.n_envs * self.n_steps, 1]) * k
        output_grads = -g / (self.n_envs * self.n_steps)
        action_grads = tape.gradient(
            action_probs, self.model.trainable_variables, output_grads
        )
        value_grads = tape.gradient(value_loss, self.model.trainable_variables)
        return [self.add_grads(g1, g2) for (g1, g2) in zip(action_grads, value_grads)]

    def update_gradients(
        self,
        states,
        rewards,
        actions,
        dones,
        previous_action_probs,
    ):
        """
        Perform gradient updates.
        Args:
            states: A tensor of shape (self.n_envs * self.n_steps, *self.input_shape)
            rewards: A tensor of shape (self.n_envs * self.n_steps)
            actions: A tensor of shape (self.n_envs * self.n_steps)
            dones: A tensor of shape (self.n_envs * self.n_steps)
            previous_action_probs: A tensor of shape (self.n_envs * self.n_steps, self.n_actions)

        Returns:
            None
        """
        action_indices = tf.concat(
            (self.batch_indices, tf.cast(actions[:, tf.newaxis], tf.int64)), -1
        )
        with tf.GradientTape(True) as tape:
            *_, critic_logits, _, action_probs = self.get_model_outputs(
                states, self.model
            )
            *_, avg_action_probs = self.get_model_outputs(states, self.avg_model)
            values = tf.reduce_sum(action_probs * critic_logits, axis=-1)
            action_probs = self.clip_last_step(action_probs)
            avg_action_probs = self.clip_last_step(avg_action_probs)
            critic_logits = self.clip_last_step(critic_logits)
            selected_probs = tf.gather_nd(action_probs, action_indices)
            selected_critic_logits = tf.gather_nd(critic_logits, action_indices)
            importance_weights = action_probs / (previous_action_probs + self.epsilon)
            selected_importance = tf.gather_nd(importance_weights, action_indices)
            returns = self.calculate_returns(
                rewards, dones, values, selected_critic_logits, selected_importance
            )
            losses = self.calculate_losses(
                action_probs,
                values,
                returns,
                selected_probs,
                selected_importance,
                selected_critic_logits,
            )
        grads = self.calculate_grads(tape, losses, action_probs, avg_action_probs)
        if self.grad_norm is not None:
            grads, norm_grads = tf.clip_by_global_norm(grads, self.grad_norm)
        self.model.optimizer.apply_gradients(zip(grads, self.model.trainable_variables))
        self.ema.apply(self.model.trainable_variables)
        tf.numpy_function(self.update_avg_weights, [], [])

    def set_batch_shapes(self, batch):
        """
        Set unknown shape batch items to their appropriate shapes,
        which for some reason, tf doesn't keep track of.
        Args:
            batch: A list of unknown shape tensors which shapes are expected
                to match self.batch_shapes.

        Returns:
            None
        """
        for item, shape in zip(batch, self.batch_shapes):
            item.set_shape(shape)

    @tf.function
    def train_step(self):
        """
        Perform 1 step which controls action_selection, interaction with environments
        in self.envs, batching and gradient updates.

        Returns:
            None
        """
        batch = tf.numpy_function(self.get_batch, [], self.batch_dtypes)
        self.buffer_current_size.assign_add(1)
        self.set_batch_shapes(batch)
        self.update_gradients(*batch)
        if (
            self.replay_ratio > 0
            and self.buffer_current_size >= self.buffers[0].initial_size
        ):
            for _ in range(np.random.poisson(self.replay_ratio)):
                batch = tf.numpy_function(
                    self.concat_buffer_samples,
                    [],
                    self.batch_dtypes,
                )
                self.set_batch_shapes(batch)
                self.update_gradients(*batch)



import os
import random
import uuid
from pathlib import Path

import gym
import pytest

from xagents.utils.common import AtariWrapper, create_buffers, get_wandb_key


@pytest.mark.parametrize(
    'resize_shape',
    [[random.randint(50, 100)] * 2 for _ in range(5)],
)
def test_atari_wrapper(resize_shape):
    """
    Test atari frame preprocessing, ensure shape and values are as expected.
    Args:
        resize_shape: (m, n) output frame size.
    """
    env = gym.make('PongNoFrameskip-v4')
    env = AtariWrapper(env, resize_shape=resize_shape)
    reset_state = env.reset()
    state, *_ = env.step(env.action_space.sample())
    assert state.shape == reset_state.shape == (*resize_shape, 1)


def test_get_wandb_key(tmp_path):
    """
    Test wandb API key is fetched correctly.
    Args:
        tmp_path: pathlib.PosixPath

    Returns:
        Wandb API key if found or None
    """
    env_var = 'WANDB_API_KEY'
    if not os.environ.get(env_var) and not (Path.home() / '.netrc').exists():
        assert not get_wandb_key()
    expected_key = uuid.uuid4().hex
    cfg_file = tmp_path / '.netrc'
    with open(cfg_file, 'w') as dummy_cfg:
        dummy_cfg.write(f'login user\npassword {expected_key}')
    assert get_wandb_key(cfg_file) == expected_key
    non_existent = tmp_path / 'non-existent'
    os.environ[env_var] = expected_key
    assert get_wandb_key(non_existent) == expected_key
    del os.environ[env_var]


@pytest.mark.parametrize(
    'max_size, batch_size, n_envs, initial_size, as_total',
    [
        [10000, 32, 16, None, True],
        [200000, 16, 16, 10000, False],
        [100000, 64, 2, None, False],
        [50000, 8, 2, 1000, True],
        [50000, 8, 3, 1000, False],
        [273645, 12, 1, 1, True],
    ],
)
def test_create_buffers(max_size, batch_size, n_envs, initial_size, as_total, agent_id):
    """
    Assert size parameters and buffer types are correct.
    Args:
        max_size: Buffer max size.
        batch_size: Buffer batch size when get_sample() is called.
        n_envs: Number of environments which will result in an equal
            number of buffers.
        initial_size: Buffer initial pre-training fill size.
        as_total: If False, total buffer initial, buffer max, and batch sizes
            are respectively buffer initial x n_envs, buffer max x n_envs,
            and batch_size x n_envs.
        agent_id: str, one of the keys in xagents.agents
    """
    expected_max_size = max_size
    expected_initial_size = initial_size or expected_max_size
    expected_batch_size = batch_size
    if as_total:
        expected_max_size //= n_envs
        expected_initial_size //= n_envs
        expected_batch_size //= n_envs
    if agent_id == 'acer':
        expected_batch_size = 1
    buffers = create_buffers(
        agent_id,
        max_size,
        batch_size,
        n_envs,
        initial_size,
        as_total,
    )
    expected_sizes = expected_max_size, expected_initial_size, expected_batch_size
    actual_sizes = buffers[0].size, buffers[0].initial_size, buffers[0].batch_size
    assert expected_sizes == actual_sizes



import configparser
import os
import re
from collections import deque
from pathlib import Path

import cv2
import gym
import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
from gym.spaces import Box, Discrete
from matplotlib import pyplot as plt
from tensorflow.keras.initializers import GlorotUniform, Orthogonal
from tensorflow.keras.layers import Conv2D, Dense, Flatten, Input
from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import Adam

import xagents
from xagents.utils.buffers import ReplayBuffer1, ReplayBuffer2


class LazyFrames:
    """
    Efficient atari frame wrapper.
    """

    def __init__(self, frames):
        """
        Wrap frames.
        Args:
            frames: numpy array of atari frames.
        """
        self.frames = frames
        self.out = None
        self.dtype = frames.dtype
        self.shape = frames.shape

    def process_frame(self):
        """
        Get wrapped frames as numpy array.
        Returns:
            frames as numpy array
        """
        if self.out is None:
            self.out = np.array(self.frames)
            self.frames = None
        return self.out

    def __array__(self, dtype=None):
        out = self.process_frame()
        if dtype is not None:
            out = out.astype(dtype)
        return out

    def __len__(self):
        return len(self.process_frame())

    def __getitem__(self, i):
        return self.process_frame()[i]

    def count(self):
        frames = self.process_frame()
        return frames.shape[frames.ndim - 1]


class AtariWrapper(gym.Wrapper):
    """
    gym wrapper for preprocessing atari frames.
    """

    def __init__(
        self,
        env,
        frame_skips=4,
        resize_shape=(84, 84),
        max_frame=False,
    ):
        """
        Initialize preprocessing settings.
        Args:
            env: gym environment that returns states as atari frames.
            frame_skips: Number of frame skips to use per environment step.
            resize_shape: (m, n) output frame size.
            max_frame: If True, max and skip is applied.
        """
        assert frame_skips > 1, 'frame_skips must be >= 1'
        super(AtariWrapper, self).__init__(env)
        self.skips = frame_skips
        self.frame_shape = resize_shape
        self.observation_space.shape = (*resize_shape, 1)
        self.max_frame = max_frame
        self.frame_buffer = deque(maxlen=2)

    def process_frame(self, frame):
        """
        Resize and convert atari frame to grayscale.
        Args:
            frame: Atari frame as numpy.ndarray

        Returns:
            Processed frame.
        """
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        frame = cv2.resize(frame, self.frame_shape)
        return LazyFrames(np.expand_dims(frame, -1))

    def step(self, action):
        """
        Step respective to self.skips.
        Args:
            action: Action supported by self.env

        Returns:
            (state, reward, done, info)
        """
        total_reward = 0
        state, done, info = 3 * [None]
        for _ in range(self.skips):
            state, reward, done, info = self.env.step(action)
            if self.max_frame:
                self.frame_buffer.append(state)
                state = np.max(np.stack(self.frame_buffer), axis=0)
            total_reward += reward
            if done:
                break
        return self.process_frame(state), total_reward, done, info

    def reset(self, **kwargs):
        """
        Reset self.env
        Args:
            **kwargs: kwargs passed to self.env.reset()

        Returns:
            Processed atari frame.
        """
        state = self.env.reset(**kwargs)
        if self.max_frame:
            self.frame_buffer.clear()
            self.frame_buffer.append(state)
        return self.process_frame(state)


def create_envs(env_name, n=1, preprocess=True, *args, **kwargs):
    """
    Create gym environment and initialize preprocessing settings.
    Args:
        env_name: Name of the environment to be passed to gym.make()
        n: Number of environments to create.
        preprocess: If True, AtariWrapper will be used.
        *args: args to be passed to AtariWrapper
        **kwargs: kwargs to be passed to AtariWrapper

    Returns:
        A list of gym environments.
    """
    envs = [gym.make(env_name) for _ in range(n)]
    if preprocess:
        assert len(envs[0].observation_space.shape) == 3, (
            f'Cannot use AtariWrapper or --preprocess for non-atari environment '
            f'{envs[0].spec.id}, with input '
            f'shape {envs[0].observation_space.shape}'
        )
        envs = [AtariWrapper(env, *args, **kwargs) for env in envs]
    return envs


class ModelReader:
    """
    Model utility class to create basic keras models from configuration files.
    """

    def __init__(self, cfg_file, output_units, input_shape, optimizer=None, seed=None):
        """
        Initialize model parser.
        Args:
            cfg_file: Path to .cfg file having that will be created.
            output_units: A list of output units that must be of the
                same size as the number of dense layers in the configuration
                without specified units.
            input_shape: input shape passed to tf.keras.layers.Input()
            optimizer: tf.keras.optimizers.Optimizer with which the resulting
                model will be compiled.
            seed: Random seed used by layer initializers.
        """
        self.initializers = {'orthogonal': Orthogonal, 'glorot_uniform': GlorotUniform}
        self.cfg_file = cfg_file
        with open(cfg_file) as cfg:
            self.parser = configparser.ConfigParser()
            self.parser.read_file(cfg)
        self.optimizer = optimizer
        self.output_units = output_units
        self.input_shape = input_shape
        self.seed = seed
        self.output_count = 0

    def get_initializer(self, section):
        """
        Get layer initializer if specified in the configuration.
        Args:
            section: str, representing section unique name.

        Returns:
            tf.keras.initializers.Initializer
        """
        initializer_name = self.parser[section].get('initializer')
        gain = self.parser[section].get('gain')
        if self.seed is not None:
            initializer_name = initializer_name or 'glorot_uniform'
        initializer_kwargs = {'seed': self.seed}
        if gain:
            initializer_kwargs.update({'gain': float(gain)})
        initializer = self.initializers.get(initializer_name)
        if initializer:
            return initializer(**initializer_kwargs)

    def create_convolution(self, section):
        """
        Parse convolution layer parameters and create layer.
        Args:
            section: str, representing section unique name.

        Returns:
            tf.keras.layers.Conv2D
        """
        filters = int(self.parser[section]['filters'])
        kernel_size = int(self.parser[section]['size'])
        stride = int(self.parser[section]['stride'])
        activation = self.parser[section].get('activation')
        return Conv2D(
            filters,
            kernel_size,
            stride,
            activation=activation,
            kernel_initializer=self.get_initializer(section),
        )

    def create_dense(self, section):
        """
        Parse dense layer parameters and create layer.
        Args:
            section: str, representing section unique name.

        Returns:
            tf.keras.layers.Dense
        """
        units = self.parser[section].get('units')
        if not units:
            assert (
                len(self.output_units) > self.output_count
            ), 'Output units given are less than dense layers required'
            units = self.output_units[self.output_count]
            self.output_count += 1
        activation = self.parser[section].get('activation')
        return Dense(
            units, activation, kernel_initializer=self.get_initializer(section)
        )

    def build_model(self):
        """
        Parse all configuration sections, create respective layers, create and
        compile model.

        Returns:
            tf.keras.Model
        """
        outputs = []
        common_layer = None
        input_layer = current_layer = Input(self.input_shape)
        sections = self.parser.sections()
        assert sections, f'Empty model configuration {self.cfg_file}'
        for section in self.parser.sections():
            if section.startswith('convolutional'):
                current_layer = self.create_convolution(section)(current_layer)
            if section.startswith('flatten'):
                current_layer = Flatten()(current_layer)
            if section.startswith('dense'):
                current_layer = self.create_dense(section)(
                    common_layer if common_layer is not None else current_layer
                )
            if self.parser[section].get('common'):
                common_layer = current_layer
            if self.parser[section].get('output'):
                outputs.append(current_layer)
        self.output_count = 0
        model = Model(input_layer, outputs)
        if self.optimizer:
            model.compile(self.optimizer)
        return model


def allocate_by_network(available_cfg, cfg_group):
    """
    Allocate given cfg file path into given group's `cnn` or `ann`
    Args:
        available_cfg: Path to .cfg file in `models` configuration folder.
        cfg_group: A dictionary with `cnn` and `ann` as keys.

    Returns:
        None
    """
    if 'cnn' in available_cfg:
        cfg_group['cnn'].append(available_cfg)
    if 'ann' in available_cfg:
        cfg_group['ann'].append(available_cfg)


def register_models(agents):
    """
    Register default model configuration files found in all agent `models`
    configuration folders to be added to xagents.agents.
    Args:
        agents: xagents.agents

    Returns:
        None
    """
    for agent_data in agents.values():
        models_folder = Path(agent_data['module'].__file__).parent / 'models'
        available_cfgs = [model_cfg.as_posix() for model_cfg in models_folder.iterdir()]
        actor_cfgs = {'cnn': [], 'ann': []}
        critic_cfgs = {'cnn': [], 'ann': []}
        model_cfgs = {'cnn': [], 'ann': []}
        for available_cfg in available_cfgs:
            if 'actor' not in available_cfg and 'critic' not in available_cfg:
                allocate_by_network(available_cfg, model_cfgs)
            elif 'actor' in available_cfg and 'critic' in available_cfg:
                allocate_by_network(available_cfg, model_cfgs)
            elif 'actor' in available_cfg:
                allocate_by_network(available_cfg, actor_cfgs)
            elif 'critic' in available_cfg:
                allocate_by_network(available_cfg, critic_cfgs)
        for key, val in zip(
            ['actor_model', 'critic_model', 'model'],
            [actor_cfgs, critic_cfgs, model_cfgs],
        ):
            if any(val.values()):
                agent_data[key] = val


def get_wandb_key(configuration_file=None):
    """
    Check ~/.netrc and WANDB_API_KEY environment variable for wandb api key.
    Args:
        configuration_file: Path to wandb configuration file, if not specified,
            defaults to ~/.netrc

    Returns:
        Key found or None
    """
    login_file = (
        Path(configuration_file) if configuration_file else Path.home() / '.netrc'
    )
    if login_file.exists():
        with open(login_file) as cfg:
            contents = cfg.read()
            key = re.findall(r"([a-fA-F\d]{32,})", contents)
            if key:
                return key[0]
    return os.environ.get('WANDB_API_KEY')


def plot_history(
    paths,
    agents,
    env,
    plot='mean_reward',
    benchmark='step',
    history_interval=1,
    time_unit='hour',
):
    """
    Plot a single agent training history given a column name,
    OR
    plot multiple agent training histories compared in a single figure.
    Args:
        paths: List of paths to .parquet files that have the training histories.
        agents: List of agent ids for labels / legend.
        env: Environment id.
        plot: Name of the column to be compared.
        benchmark: `step` or `time`
        history_interval: Plot every n data points.
        time_unit: Assuming the `time` column has seconds, time
            will be divided accordingly.

    Returns:
        None
    """
    time_divisors = {'hour': 3600, 'minute': 60, 'second': 1}
    assert len(paths) == len(agents), (
        f'Expected `paths` and `agents` to have the same sizes, '
        f'got {len(paths)} vs {len(agents)}'
    )
    histories = [
        pd.read_parquet(path).sort_values('time').iloc[::history_interval]
        for path in paths
    ]
    for history in histories:
        data = history[benchmark]
        if benchmark == 'time':
            data = history[benchmark] / time_divisors[time_unit]
        plt.plot(data, history[plot])
    if len(agents) == 1:
        title = f'{agents[0]} - {env}'
    else:
        title = env
    plt.title(title)
    x_label = f'{benchmark} ({time_unit}s)' if benchmark == 'time' else benchmark
    plt.xlabel(x_label)
    plt.ylabel(plot.replace('_', ' '))
    plt.legend(agents, loc='lower right')
    plt.grid()


def write_from_dict(_dict, path):
    """
    Write to .parquet given a dict
    Args:
        _dict: Dictionary of label: [scalar]
        path: Path to .parquet fiile.

    Returns:
        None
    """
    table = pa.Table.from_pydict(_dict)
    pq.write_to_dataset(table, root_path=path, compression='gzip')


def create_model(
    env, agent_id, model_type, optimizer_kwargs=None, seed=None, model_cfg=None
):
    """
    Create model from a given model cfg or from default configuration if any.
    Args:
        env: gym environment.
        agent_id: str, one of the keys in xagents.agents
        model_type: 'model' or 'actor_model' or 'critic_model'
        optimizer_kwargs: A dictionary of epsilon, that may contain one or more of
            `learning_rate`, `beta_1` or `beta_2`
        seed: random seed passed to layer initializers.
        model_cfg: Path to .cfg containing a compatible model configuration.

    Returns:
        tf.keras.Model
    """
    units = [
        env.action_space.n
        if isinstance(env.action_space, Discrete)
        else env.action_space.shape[0]
    ]
    if len(env.observation_space.shape) == 3:
        network_type = 'cnn'
    else:
        network_type = 'ann'
    try:
        model_cfg = model_cfg or xagents.agents[agent_id][model_type][network_type][0]
    except IndexError:
        model_cfg = None
    models_folder = Path(xagents.agents[agent_id]['module'].__file__).parent / 'models'
    assert model_cfg, (
        f'You should specify `model_cfg`. No default '
        f'{network_type.upper()} model found in\n{models_folder}'
    )
    if agent_id == 'acer':
        units.append(units[-1])
    elif 'actor' in model_cfg and 'critic' in model_cfg:
        units.append(1)
    elif 'critic' in model_cfg:
        units[0] = 1
    optimizer_kwargs = optimizer_kwargs or {}
    model_reader = ModelReader(
        model_cfg,
        units,
        env.observation_space.shape,
        Adam(**optimizer_kwargs),
        seed,
    )
    if agent_id in ['td3', 'ddpg'] and 'critic' in model_cfg:
        assert isinstance(env.action_space, Box), (
            f'Invalid environment: {env.spec.id}. {agent_id.upper()} supports '
            f'environments with a Box action space only, got {env.action_space}'
        )
        model_reader.input_shape = (
            model_reader.input_shape[0] + env.action_space.shape[0]
        )
    return model_reader.build_model()


def create_models(options, env, agent_id, **kwargs):
    """
    Create agent models
    Args:
        options: iterable that has `model` or (`actor_model` and `critic_model`)
        env: gym environment
        agent_id: str, one of the keys in xagents.agents
        **kwargs: kwargs passed to xagents.utils.common.create_model()

    Returns:
        list of model(s)
    """
    model_types = ['model', 'actor_model', 'critic_model']
    models = {}
    for model_type in model_types:
        if model_type in options:
            model_cfg = options[model_type]
            if not isinstance(model_cfg, (str, Path)):
                model_cfg = None
            models[model_type] = create_model(
                env, agent_id, model_type, model_cfg=model_cfg, **kwargs
            )
    return models


def create_buffers(
    agent_id,
    max_size,
    batch_size,
    n_envs,
    initial_size=None,
    as_total=True,
):
    """
    Create deque-based or numpy-based replay buffers.
    Args:
        agent_id: str, one of the keys in xagents.agents
        max_size: Buffer max size.
        batch_size: Buffer batch size when get_sample() is called.
        n_envs: Number of environments which will result in an equal
            number of buffers.
        initial_size: Buffer initial pre-training fill size.
        as_total: If False, total buffer initial, buffer max, and batch sizes
            are respectively buffer initial x n_envs, buffer max x n_envs,
            and batch_size x n_envs.

    Returns:
        list of buffers.
    """
    initial_size = initial_size or max_size
    if as_total:
        max_size //= n_envs
        initial_size //= n_envs
        batch_size //= n_envs
    if agent_id == 'acer':
        batch_size = 1
    if agent_id in ['td3', 'ddpg']:
        buffers = [
            ReplayBuffer2(
                max_size,
                5,
                initial_size=initial_size,
                batch_size=batch_size,
            )
            for _ in range(n_envs)
        ]
    else:
        buffers = [
            ReplayBuffer1(
                max_size,
                initial_size=initial_size,
                batch_size=batch_size,
            )
            for _ in range(n_envs)
        ]
    return buffers


def create_agent(agent_id, agent_kwargs, non_agent_kwargs, trial=None):
    """
    Create agent with all sub-components, including environments,
    models and buffers.
    Args:
        agent_id: str, one of the keys in xagents.agents
        agent_kwargs: dictionary of agent kwargs, values
        non_agent_kwargs: dictionary of non-agent kwargs, values
        trial: optuna.trial.Trial

    Returns:
        agent.
    """
    agent_kwargs['trial'] = trial
    envs = create_envs(
        non_agent_kwargs['env'],
        non_agent_kwargs['n_envs'],
        non_agent_kwargs['preprocess'],
        max_frame=non_agent_kwargs['max_frame'],
    )
    agent_kwargs['envs'] = envs
    optimizer_kwargs = {
        'learning_rate': non_agent_kwargs['lr'],
        'beta_1': non_agent_kwargs['beta1'],
        'beta_2': non_agent_kwargs['beta2'],
        'epsilon': non_agent_kwargs['opt_epsilon'],
    }
    models = create_models(
        agent_kwargs,
        envs[0],
        agent_id,
        optimizer_kwargs=optimizer_kwargs,
        seed=agent_kwargs['seed'],
    )
    agent_kwargs.update(models)
    if (
        issubclass(xagents.agents[agent_id]['agent'], xagents.OffPolicy)
        or agent_id == 'acer'
    ):
        buffers = create_buffers(
            agent_id,
            non_agent_kwargs['buffer_max_size'],
            non_agent_kwargs['buffer_batch_size'],
            non_agent_kwargs['n_envs'],
            non_agent_kwargs['buffer_initial_size'],
        )
        agent_kwargs['buffers'] = buffers
    agent = xagents.agents[agent_id]['agent'](**agent_kwargs)
    if non_agent_kwargs['weights']:
        n_weights = len(non_agent_kwargs['weights'])
        n_models = len(agent.output_models)
        assert (
            n_weights == n_models
        ), f'Expected {n_models} weights to load, got {n_weights}'
        for weight, model in zip(non_agent_kwargs['weights'], agent.output_models):
            model.load_weights(weight).expect_partial()
    return agent



cli_args = {
    'model': {'help': 'Path to model .cfg file'},
    'entropy-coef': {
        'help': 'Entropy coefficient for loss calculation',
        'type': float,
        'default': 0.01,
        'hp_type': 'log_uniform',
    },
    'value-loss-coef': {
        'help': 'Value loss coefficient for value loss calculation',
        'type': float,
        'default': 0.5,
        'hp_type': 'log_uniform',
    },
    'grad-norm': {
        'help': 'Gradient clipping value passed to tf.clip_by_value()',
        'type': float,
        'default': 0.5,
        'hp_type': 'log_uniform',
    },
    'n-steps': {
        'help': 'Transition steps',
        'type': int,
        'default': 5,
        'hp_type': 'categorical',
    },
}



import numpy as np
import tensorflow as tf

from xagents import A2C


class PPO(A2C):
    """
    Proximal Policy Optimization Algorithms.
    https://arxiv.org/abs/1707.06347
    """

    def __init__(
        self,
        envs,
        model,
        lam=0.95,
        ppo_epochs=4,
        mini_batches=4,
        advantage_epsilon=1e-8,
        clip_norm=0.1,
        **kwargs,
    ):
        """
        Initialize PPO agent.
        Args:
            envs: A list of gym environments.
            model: tf.keras.models.Model that is expected to be compiled
                with an optimizer before training starts.
            lam: GAE-Lambda for advantage estimation
            ppo_epochs: Gradient updates per training step.
            mini_batches: Number of mini batches to use per gradient update.
            advantage_epsilon: Epsilon value added to estimated advantage.
            clip_norm: Clipping value passed to tf.clip_by_value()
            **kwargs: kwargs Passed to super classes.
        """
        super(PPO, self).__init__(envs, model, **kwargs)
        self.lam = lam
        self.ppo_epochs = ppo_epochs
        self.mini_batches = mini_batches
        self.advantage_epsilon = advantage_epsilon
        self.clip_norm = clip_norm
        self.batch_size = self.n_envs * self.n_steps
        self.mini_batch_size = self.batch_size // self.mini_batches
        assert (
            self.mini_batch_size > 0
        ), f'Invalid batch size to mini-batch size ratio {self.batch_size}: {self.mini_batches}'

    def calculate_returns(
        self,
        rewards,
        dones,
        values=None,
        selected_critic_logits=None,
        selected_importance=None,
    ):
        """
        Get a batch of returns.
        Args:
            rewards: Rewards tensor of shape (self.n_steps, self.n_envs)
            dones: Dones tensor of shape (self.n_steps, self.n_envs)
            values: Values tensor of shape (self.n_steps + 1, self.n_envs).
                required for PPO, TRPO and ACER
            selected_critic_logits: Critic output respective to selected actions
                of shape (self.n_steps, self.n_envs).
                Required for ACER.
            selected_importance: Importance weights respective to selected
                actions of shape (self.n_steps, self.n_envs).
                Required for ACER
        Returns:
            Tensor of n-step returns.
        """
        next_values = self.get_model_outputs(self.get_states(), self.output_models)[
            2
        ].numpy()
        next_values = (
            np.expand_dims(next_values, 0)
            if not len(next_values.shape)
            else next_values
        )
        returns = []
        last_lam = 0
        values = np.concatenate([values, np.expand_dims(next_values, 0)])
        dones = np.concatenate([dones, np.expand_dims(dones[-1], 0)])
        for step in reversed(range(self.n_steps)):
            next_non_terminal = 1 - dones[step + 1]
            next_values = values[step + 1]
            delta = (
                rewards[step]
                + self.gamma * next_values * next_non_terminal
                - values[step]
            )
            last_lam = delta + self.gamma * self.lam * next_non_terminal * last_lam
            returns.append(last_lam)
        return np.asarray(returns[::-1]) + values[:-1]

    def update_gradients(
        self, states, actions, old_values, returns, old_log_probs, advantages
    ):
        """
        Perform gradient updates.
        Args:
            states: states as numpy array of shape (self.mini_batch_size, *self.input_shape)
            actions: actions as numpy array of shape (self.mini_batch_size,)
            old_values: old values as numpy array of shape (self.mini_batch_size,)
            returns: returns as numpy array of shape (self.mini_batch_size,)
            old_log_probs: old log probs as numpy array of shape (self.mini_batch_size,)
            advantages: advantages as numpy array of shape (self.mini_batch_size,)

        Returns:
            None
        """
        with tf.GradientTape() as tape:
            _, log_probs, values, entropy, _ = self.get_model_outputs(
                states, self.output_models, actions=actions
            )
            entropy = tf.reduce_mean(entropy)
            clipped_values = old_values + tf.clip_by_value(
                values - old_values, -self.clip_norm, self.clip_norm
            )
            value_loss1 = tf.square(values - returns)
            value_loss2 = tf.square(clipped_values - returns)
            value_loss = 0.5 * tf.reduce_mean(tf.maximum(value_loss1, value_loss2))
            ratio = tf.exp(log_probs - old_log_probs)
            pg_loss1 = -advantages * ratio
            pg_loss2 = -advantages * tf.clip_by_value(
                ratio, 1 - self.clip_norm, 1 + self.clip_norm
            )
            pg_loss = tf.reduce_mean(tf.maximum(pg_loss1, pg_loss2))
            loss = (
                pg_loss
                - entropy * self.entropy_coef
                + value_loss * self.value_loss_coef
            )
        grads = tape.gradient(loss, self.model.trainable_variables)
        if self.grad_norm is not None:
            grads, _ = tf.clip_by_global_norm(grads, self.grad_norm)
        self.model.optimizer.apply_gradients(zip(grads, self.model.trainable_variables))

    def get_mini_batches(self, *args):
        """
        Split each item in args into mini-batches of size self.mini_batch_size.
        Args:
            *args: Tensors of the same size to split.

        Returns:
            A list of mini-batches as tensors.
        """
        mini_batches = []
        indices = tf.range(self.batch_size)
        for _ in range(self.ppo_epochs):
            indices = tf.random.shuffle(indices)
            for i in range(0, self.batch_size, self.mini_batch_size):
                batch_indices = indices[i : i + self.mini_batch_size]
                mini_batches.append([tf.gather(item, batch_indices) for item in args])
        return mini_batches

    def run_ppo_epochs(self, states, actions, returns, old_values, old_log_probs):
        """
        Split batch into mini batches and perform gradient updates.
        Args:
            states: states as numpy array of shape (self.n_steps * self.n_envs, *self.input_shape)
            actions: actions as numpy array of shape (self.n_steps * self.n_envs,)
            returns: returns as numpy array of shape (self.n_steps * self.n_envs,)
            old_values: old values as numpy array of shape (self.n_steps * self.n_envs,)
            old_log_probs: old log probs as numpy array of shape (self.n_steps * self.n_envs,)

        Returns:
            None
        """
        for mini_batch in self.get_mini_batches(
            states, actions, returns, old_values, old_log_probs
        ):
            (
                states_mb,
                actions_mb,
                returns_mb,
                old_values_mb,
                old_log_probs_mb,
            ) = mini_batch
            advantages_mb = returns_mb - old_values_mb
            advantages_mb = (advantages_mb - tf.reduce_mean(advantages_mb)) / (
                tf.math.reduce_std(advantages_mb) + self.advantage_epsilon
            )
            self.update_gradients(
                states_mb,
                actions_mb,
                old_values_mb,
                returns_mb,
                old_log_probs_mb,
                advantages_mb,
            )

    def get_batch(self):
        """
        Get n-step batch which is the result of running self.envs step() for
        self.n_steps times, calculate returns and adjust n-step batch shapes
        for gradient updates.

        Returns:
            [states, actions, returns, values, log probs] with adjusted shapes.
        """
        (
            states,
            rewards,
            actions,
            values,
            dones,
            log_probs,
            *_,
        ) = [np.asarray(item, np.float32) for item in super(PPO, self).get_batch()]
        values = np.expand_dims(values, -1) if len(values.shape) <= 1 else values
        returns = self.calculate_returns(rewards, dones, values)
        return self.concat_step_batches(states, actions, returns, values, log_probs)

    @tf.function
    def train_step(self):
        """
        Perform 1 step which controls action_selection, interaction with environments
        in self.envs, batching and gradient updates.

        Returns:
            None
        """
        batch = tf.numpy_function(self.get_batch, [], 5 * [tf.float32])
        self.run_ppo_epochs(*batch)



from xagents import a2c, acer, ddpg, dqn, ppo, td3, trpo
from xagents.a2c.agent import A2C
from xagents.acer.agent import ACER
from xagents.base import OffPolicy
from xagents.ddpg.agent import DDPG
from xagents.dqn.agent import DQN
from xagents.ppo.agent import PPO
from xagents.td3.agent import TD3
from xagents.trpo.agent import TRPO
from xagents.utils.cli import play_args, train_args, tune_args
from xagents.utils.common import register_models

__author__ = 'schissmantics'
__email__ = 'schissmantics@outlook.com'
__license__ = 'MIT'
__version__ = '1.0.1'

agents = {
    'a2c': {'module': a2c, 'agent': A2C},
    'acer': {'module': acer, 'agent': ACER},
    'dqn': {'module': dqn, 'agent': DQN},
    'ppo': {'module': ppo, 'agent': PPO},
    'td3': {'module': td3, 'agent': TD3},
    'trpo': {'module': trpo, 'agent': TRPO},
    'ddpg': {'module': ddpg, 'agent': DDPG},
}
register_models(agents)
commands = {
    'train': (train_args, 'fit', 'Train given an agent and environment'),
    'play': (
        play_args,
        'play',
        'Play a game given a trained agent and environment',
    ),
    'tune': (
        tune_args,
        '',
        'Tune hyperparameters given an agent, hyperparameter specs, and environment',
    ),
}



import numpy as np
import tensorflow as tf
from gym.spaces.discrete import Discrete
from tensorflow.keras.losses import MSE

from xagents.base import OffPolicy


class DQN(OffPolicy):
    """
    Playing Atari with Deep Reinforcement Learning.
    https://arxiv.org/abs/1312.5602
    """

    def __init__(
        self,
        envs,
        model,
        buffers,
        double=False,
        epsilon_start=1.0,
        epsilon_end=0.02,
        epsilon_decay_steps=150000,
        target_sync_steps=1000,
        **kwargs,
    ):
        """
        Initialize DQN agent.
        Args:
            envs: A list of gym environments.
            model: tf.keras.models.Model that is expected to be compiled
                with an optimizer before training starts.
            buffers: A list of replay buffer objects whose length should match
                `envs`s'.
            double: If True, DDQN is used.
            epsilon_start: Starting epsilon value which is used to control random exploration.
                It should be decremented and adjusted according to implementation needs.
            epsilon_end: End epsilon value which is the minimum exploration rate.
            epsilon_decay_steps: Number of steps for epsilon to reach `epsilon_end`
                from `epsilon_start`,
            target_sync_steps: Steps to sync target models after each.
            **kwargs: kwargs passed to super classes.
        """
        super(DQN, self).__init__(envs, model, buffers, **kwargs)
        self.assert_valid_env(envs[0], Discrete)
        self.target_model = tf.keras.models.clone_model(self.model)
        self.target_model.set_weights(self.model.get_weights())
        self.double = double
        self.epsilon_start = self.epsilon = epsilon_start
        self.epsilon_end = epsilon_end
        self.epsilon_decay_steps = epsilon_decay_steps
        self.target_sync_steps = target_sync_steps
        self.batch_indices = tf.range(
            self.buffers[0].batch_size * self.n_envs, dtype=tf.int64
        )[:, tf.newaxis]
        self.batch_dtypes = ['uint8', 'int64', 'float64', 'bool', 'uint8']

    @staticmethod
    def get_action_indices(batch_indices, actions):
        """
        Get indices that will be passed to tf.gather_nd()
        Args:
            batch_indices: tf.range() result of the same shape as the batch size.
            actions: Action tensor of same shape as the batch size.

        Returns:
            Indices as a tensor.
        """
        return tf.concat((batch_indices, tf.cast(actions[:, tf.newaxis], tf.int64)), -1)

    @tf.function
    def get_model_outputs(self, inputs, models, training=True):
        """
        Get model outputs (actions)
        Args:
            inputs: Inputs as tensors / numpy arrays that are expected
                by the given model.
            models: A tf.keras.Model or a list of tf.keras.Model(s)
            training: `training` parameter passed to model call.

        Returns:
            Outputs that is expected from the given model.
        """
        q_values = super(DQN, self).get_model_outputs(inputs, models, training)
        return tf.argmax(q_values, 1), q_values

    def update_epsilon(self):
        """
        Decrement epsilon which aims to gradually reduce randomization.

        Returns:
            None
        """
        self.epsilon = max(
            self.epsilon_end, self.epsilon_start - self.steps / self.epsilon_decay_steps
        )

    def sync_target_model(self):
        """
        Sync target model weights with main's

        Returns:
            None
        """
        if self.steps % self.target_sync_steps == 0:
            self.target_model.set_weights(self.model.get_weights())

    def get_actions(self):
        """
        Generate action following an epsilon-greedy policy.

        Returns:
            A random action or Q argmax.
        """
        if np.random.random() < self.epsilon:
            return np.random.randint(0, self.n_actions, self.n_envs)
        return self.get_model_outputs(self.get_states(), self.output_models)[0]

    def get_targets(self, states, actions, rewards, dones, new_states):
        """
        Get targets for gradient update.
        Args:
            states: A tensor of shape (self.n_envs * total buffer batch size, *self.input_shape)
            actions: A tensor of shape (self.n_envs * total buffer batch size)
            rewards: A tensor of shape (self.n_envs * total buffer batch size)
            dones: A tensor of shape (self.n_envs * total buffer batch size)
            new_states: A tensor of shape (self.n_envs * total buffer batch size, *self.input_shape)

        Returns:
            Target values, a tensor of shape (self.n_envs * total buffer batch size, self.n_actions)
        """
        q_states = self.get_model_outputs(states, self.model)[1]
        if self.double:
            new_state_actions = self.get_model_outputs(new_states, self.model)[0]
            new_state_q_values = self.get_model_outputs(new_states, self.target_model)[
                1
            ]
            a = self.get_action_indices(self.batch_indices, new_state_actions)
            new_state_values = tf.gather_nd(new_state_q_values, a)
        else:
            new_state_values = tf.reduce_max(
                self.get_model_outputs(new_states, self.target_model)[1], axis=1
            )
        new_state_values = tf.where(
            tf.cast(dones, tf.bool),
            tf.constant(0, new_state_values.dtype),
            new_state_values,
        )
        target_values = tf.identity(q_states)
        target_value_update = new_state_values * self.gamma + tf.cast(
            rewards, tf.float32
        )
        indices = self.get_action_indices(self.batch_indices, actions)
        target_values = tf.tensor_scatter_nd_update(
            target_values, indices, target_value_update
        )
        return target_values

    def update_gradients(self, x, y):
        """
        Train on a given batch.
        Args:
            x: States tensor
            y: Targets tensor

        Returns:
            None
        """
        with tf.GradientTape() as tape:
            y_pred = self.get_model_outputs(x, self.model)[1]
            loss = MSE(y, y_pred)
        self.model.optimizer.minimize(loss, self.model.trainable_variables, tape=tape)

    def at_step_start(self):
        """
        Execute steps that will run before self.train_step() which decays epsilon.

        Returns:
            None
        """
        self.update_epsilon()

    @tf.function
    def train_step(self):
        """
        Perform 1 step which controls action_selection, interaction with environments
        in self.envs, batching and gradient updates.

        Returns:
            None
        """
        actions = tf.numpy_function(self.get_actions, [], tf.int64)
        tf.numpy_function(self.step_envs, [actions, False, True], [])
        training_batch = tf.numpy_function(
            self.concat_buffer_samples,
            [],
            self.batch_dtypes,
        )
        targets = self.get_targets(*training_batch)
        self.update_gradients(training_batch[0], targets)

    def at_step_end(self):
        """
        Execute steps that will run after self.train_step() which
        updates target model.

        Returns:
            None
        """
        self.sync_target_model()



import gym
import numpy as np
import optuna
import pytest
from tensorflow.keras.layers import Dense, Input
from tensorflow.keras.models import Model

import xagents
from xagents import ACER, DDPG, DQN, TD3
from xagents.base import BaseAgent, OffPolicy, OnPolicy
from xagents.cli import Executor
from xagents.utils.buffers import ReplayBuffer1, ReplayBuffer2
from xagents.utils.cli import (agent_args, non_agent_args, play_args,
                               train_args, tune_args)

optuna.logging.set_verbosity(optuna.logging.ERROR)


@pytest.fixture(scope='function')
def executor(request):
    """
    Fixture for testing command line options.
    Args:
        request: _pytest.fixtures.SubRequest

    Returns:
        xagents.cli.Executor
    """
    request.cls.executor = Executor()


@pytest.fixture(
    params=[
        ('non-agent', non_agent_args),
        ('agent', agent_args),
        ('train', train_args),
        ('play', play_args),
        ('tune', tune_args),
        ('do-nothing', {}),
    ]
)
def section(request):
    """
    Fixture for testing help menu and parsing sanity.
    Args:
        request: _pytest.fixtures.SubRequest

    Yields:
        Tuple of argument group name and respective dict of args.
    """
    yield request.param


@pytest.fixture(scope='class')
def envs(request):
    """
    Fixture of PongNoFrameskip-v4 environments used to test agents that support
    environments with Discrete action space.
    Args:
        request: _pytest.fixtures.SubRequest

    Returns:
        A list of 4 pong environments.
    """
    request.cls.envs = [gym.make('PongNoFrameskip-v4') for _ in range(4)]


@pytest.fixture(scope='class')
def envs2(request):
    """
    Fixture of BipedalWalker-v3 environments used to test agents that support
    environments with continuous action space.
    Args:
        request: _pytest.fixtures.SubRequest

    Returns:
        A list of 4 bipedal walker environments.
    """
    request.cls.envs2 = [gym.make('BipedalWalker-v3') for _ in range(4)]


@pytest.fixture(scope='class')
def model(request):
    """
    Fixture used by all agents in tests that require a model but do not
    require the agent default model. It should be included in
    pytest.mark.usefixtures() decorator to allow its availability as
    a class attribute.
    Args:
        request: _pytest.fixtures.SubRequest

    Returns:
        None
    """
    x0 = Input((210, 160, 3))
    x = Dense(1, 'relu')(x0)
    x = Dense(1, 'relu')(x)
    x = Dense(1, 'relu')(x)
    model = request.cls.model = Model(x0, x)
    model.compile('adam')


@pytest.fixture(scope='class')
def buffers(request):
    """
    Fixture used by off-policy agent tests that require a replay buffer
    that is not size specific. It should be included in
    pytest.mark.usefixtures() decorator to allow its availability as
    a class attribute.
    Args:
        request: _pytest.fixtures.SubRequest

    Returns:
        None
    """
    request.cls.buffers = [ReplayBuffer1(200, batch_size=155) for _ in range(4)]


@pytest.fixture(scope='function')
def buffer1(request):
    """
    Fixture used to test ReplayBuffer1.
    Args:
        request: _pytest.fixtures.SubRequest

    Returns:
        None
    """
    request.cls.buffer = ReplayBuffer1(100, batch_size=4)


@pytest.fixture(scope='function')
def buffer2(request):
    """
    Fixture used to test ReplayBuffer2.
    Args:
        request: _pytest.fixtures.SubRequest

    Returns:
        None
    """
    request.cls.buffer = ReplayBuffer2(100, 5, batch_size=4)


@pytest.fixture(params=[item['agent'] for item in xagents.agents.values()])
def agent(request):
    """
    Fixture with agents available in xagents.agents
    Args:
        request: _pytest.fixtures.SubRequest

    Yields:
        OnPolicy/OffPolicy subclass.
    """
    yield request.param


@pytest.fixture(params=[agent_id for agent_id in xagents.agents])
def agent_id(request):
    """
    Fixture with agent ids available in xagents.agents
    Args:
        request: _pytest.fixtures.SubRequest

    Yields:
        Agent id as str
    """
    yield request.param


@pytest.fixture(params=[command for command in xagents.commands])
def command(request):
    """
    Fixture with commands available in xagents.commands
    Args:
        request: _pytest.fixtures.SubRequest

    Yields:
        Command as str
    """
    yield request.param


@pytest.fixture(params=[ACER, TD3, DQN, DDPG])
def off_policy_agent(request):
    """
    Fixture with off-policy agents.
    Args:
        request: _pytest.fixtures.SubRequest

    Yields:
        OffPolicy subclass.
    """
    yield request.param


@pytest.fixture(params=[BaseAgent, OnPolicy, OffPolicy])
def base_agent(request):
    """
    Fixture with agent base classes, used to test base and
    abstract methods.
    Args:
        request: _pytest.fixtures.SubRequest

    Yields:
        OnPolicy/OffPolicy subclass.
    """
    yield request.param


@pytest.fixture(params=[ReplayBuffer1, ReplayBuffer2])
def buffer_type(request):
    """
    Fixture with buffer types to test buffer attributes.
    Args:
        request: _pytest.fixtures.SubRequest

    Yields:
        BaseBuffer subclass.
    """
    yield request.param


@pytest.fixture(scope='class')
def observations(request):
    """
    Fixture to be used by class methods that require creation of a batch.
    Args:
        request: _pytest.fixtures.SubRequest

    Returns:
        None
    """
    states = np.random.randint(0, 255, (10, 84, 84, 1))
    actions = np.random.randint(0, 6, 10)
    rewards = np.random.randint(0, 10, 10)
    dones = np.random.randint(0, 1000000, 10)
    new_states = np.random.randint(0, 255, (10, 84, 84, 1))
    request.cls.observations = [
        [*items] for items in zip(states, actions, rewards, dones, new_states)
    ]


@pytest.fixture(scope='class')
def study(request):
    """
    Fixture that provides optuna.study.Study to test tuning components.
    Args:
        request: _pytest.fixtures.SubRequest

    Returns:
        None
    """
    pruner = optuna.pruners.MedianPruner(1)
    request.cls.study = optuna.create_study(pruner=pruner, direction='maximize')



import xagents
from xagents.base import OffPolicy
from xagents.utils.cli import agent_args, non_agent_args, off_policy_args


def get_expected_flags(argv, as_kwargs=False):
    """
    Convert argv to expected flags/keyword that should be present
    as argparse.Namespace attributes and as kwargs that will be
    passed directly to agent.
    Args:
        argv: Arguments passed.
        as_kwargs: If True example-flag1 will be returned as example_flag1

    Returns:
        List of ['example-flag1', 'example-flag2', ...]
        or
        list of ['example_flag1', 'example_flag2']
    """
    if not argv:
        return []
    command = argv[0]
    expected_kwargs = {}
    expected_kwargs.update(agent_args)
    expected_kwargs.update(non_agent_args)
    expected_kwargs.update(xagents.commands[command][0])
    if len(argv) > 1:
        agent_data = xagents.agents[argv[1]]
        expected_kwargs.update(agent_data['module'].cli_args)
        if issubclass(agent_data['agent'], OffPolicy) or argv[1] == 'acer':
            expected_kwargs.update(off_policy_args)
    if not as_kwargs:
        return expected_kwargs.keys()
    return [flag.replace('-', '_') for flag in expected_kwargs.keys()]


def assert_flags_displayed(cap, title, cli_args):
    """
    Assert title and respective flags are present in help menu.
    Args:
        cap: str, text displayed to the console
        title: str, that should be present in cap representing
            the title of the group of argument group.
        cli_args: A dictionary of flags and their attributes.

    Returns:
        None
    """
    assert title in cap
    for flag in cli_args:
        assert f'--{flag}' in cap
        for key in cli_args[flag]:
            if key in ['help', 'hp_type', 'default']:
                for line in str(cli_args[flag][key]).split('\n'):
                    assert line in cap



import tensorflow as tf
from tensorflow.keras.losses import MSE

from xagents import DDPG


class TD3(DDPG):
    """
    Addressing Function Approximation Error in Actor-Critic Methods.
    https://arxiv.org/abs/1802.09477
    """

    def __init__(
        self,
        envs,
        actor_model,
        critic_model,
        buffers,
        policy_delay=2,
        policy_noise_coef=0.2,
        noise_clip=0.5,
        **kwargs,
    ):
        """
        Initialize TD3 agent
        Args:
            envs: A list of gym environments.
            actor_model: Actor separate model as a tf.keras.Model.
            critic_model: Critic separate model as a tf.keras.Model.
            buffers: A list of replay buffer objects whose length should match
                `envs`s'.
            policy_delay: Number of gradient steps after which, actor weights
                will be updated as well as syncing the target models weights.
            policy_noise_coef: Coefficient multiplied by noise added to target actions.
            noise_clip: Target noise clipping value.
            **kwargs: kwargs passed to super classes.
        """
        super(TD3, self).__init__(envs, actor_model, critic_model, buffers, **kwargs)
        self.critic1 = self.critic
        self.target_critic1 = self.target_critic
        self.policy_delay = policy_delay
        self.policy_noise_coef = policy_noise_coef
        self.noise_clip = noise_clip
        self.critic2 = tf.keras.models.clone_model(self.critic1)
        self.output_models.append(self.critic2)
        self.critic2.compile(
            tf.keras.optimizers.get(self.critic1.optimizer.get_config()['name'])
        )
        self.critic2.optimizer.learning_rate.assign(
            self.critic1.optimizer.learning_rate
        )
        self.target_critic2 = tf.keras.models.clone_model(self.critic1)
        self.target_critic2.set_weights(self.critic2.get_weights())
        self.model_groups.append(
            (self.critic2, self.target_critic2),
        )

    def get_step_actions(self):
        """
        Get self.n_envs actions to be stepped by self.envs

        Returns:
            actions
        """
        return self.actor(tf.numpy_function(self.get_states, [], self.states[0].dtype))

    def update_critic_weights(self, states, actions, new_states, dones, rewards):
        """
        Update critic 1 and critic 2 weights.
        Args:
            states: A tensor of shape (self.n_envs * total buffer batch size, *self.input_shape)
            actions: A tensor of shape (self.n_envs * total buffer batch size, self.n_actions)
            new_states: A tensor of shape (self.n_envs * total buffer batch size, *self.input_shape)
            dones: A tensor of shape (self.n_envs * total buffer batch size)
            rewards: A tensor of shape (self.n_envs * total buffer batch size)

        Returns:
            None
        """
        with tf.GradientTape(True) as tape:
            noise = (
                tf.random.normal(
                    (self.buffers[0].batch_size * self.n_envs, self.n_actions)
                )
                * self.policy_noise_coef
            )
            noise = tf.clip_by_value(noise, -self.noise_clip, self.noise_clip)
            new_actions = tf.clip_by_value(
                self.target_actor(new_states) + noise, -1.0, 1.0
            )
            target_critic_input = tf.concat(
                [tf.cast(new_states, tf.float64), tf.cast(new_actions, tf.float64)], 1
            )
            target_value1 = self.target_critic1(target_critic_input)
            target_value2 = self.target_critic2(target_critic_input)
            target_value = tf.minimum(target_value1, target_value2)
            target_value = rewards + tf.stop_gradient(
                (1 - dones) * self.gamma * target_value
            )
            critic_input = tf.concat([states, actions], 1)
            value1 = self.critic1(critic_input)
            value2 = self.critic2(critic_input)
            critic1_loss, critic2_loss = MSE(value1, target_value), MSE(
                value2, target_value
            )
        self.critic1.optimizer.minimize(
            critic1_loss, self.critic1.trainable_variables, tape=tape
        )
        self.critic2.optimizer.minimize(
            critic2_loss, self.critic2.trainable_variables, tape=tape
        )



non_agent_args = {
    'env': {'help': 'gym environment id', 'required': True},
    'n-envs': {
        'help': 'Number of environments to create',
        'default': 1,
        'type': int,
        'hp_type': 'categorical',
    },
    'preprocess': {
        'help': 'If specified, states will be treated as atari frames\n'
        'and preprocessed accordingly',
        'action': 'store_true',
    },
    'lr': {
        'help': 'Learning rate passed to a tensorflow.keras.optimizers.Optimizer',
        'type': float,
        'default': 7e-4,
        'hp_type': 'log_uniform',
    },
    'opt-epsilon': {
        'help': 'Epsilon passed to a tensorflow.keras.optimizers.Optimizer',
        'type': float,
        'default': 1e-7,
        'hp_type': 'log_uniform',
    },
    'beta1': {
        'help': 'Beta1 passed to a tensorflow.keras.optimizers.Optimizer',
        'type': float,
        'default': 0.9,
        'hp_type': 'log_uniform',
    },
    'beta2': {
        'help': 'Beta2 passed to a tensorflow.keras.optimizers.Optimizer',
        'type': float,
        'default': 0.999,
        'hp_type': 'log_uniform',
    },
    'weights': {
        'help': 'Path(s) to model(s) weight(s) to be loaded by agent output_models',
        'nargs': '+',
    },
    'max-frame': {
        'help': 'If specified, max & skip will be applied during preprocessing',
        'action': 'store_true',
        'hp_type': 'categorical',
    },
}

off_policy_args = {
    'buffer-max-size': {
        'help': 'Maximum replay buffer size',
        'type': int,
        'default': 10000,
        'hp_type': 'int',
    },
    'buffer-initial-size': {
        'help': 'Replay buffer initial size',
        'type': int,
        'hp_type': 'int',
    },
    'buffer-batch-size': {
        'help': 'Replay buffer batch size',
        'type': int,
        'default': 32,
        'hp_type': 'categorical',
    },
}

agent_args = {
    'reward-buffer-size': {
        'help': 'Size of the total reward buffer, used for calculating\n'
        'mean reward value to be displayed.',
        'default': 100,
        'type': int,
    },
    'gamma': {
        'help': 'Discount factor',
        'default': 0.99,
        'type': float,
        'hp_type': 'log_uniform',
    },
    'display-precision': {
        'help': 'Number of decimals to be displayed',
        'default': 2,
        'type': int,
    },
    'seed': {'help': 'Random seed', 'type': int},
    'log-frequency': {'help': 'Log progress every n games', 'type': int},
    'checkpoints': {
        'help': 'Path(s) to new model(s) to which checkpoint(s) will be saved during training',
        'nargs': '+',
    },
    'history-checkpoint': {'help': 'Path to .parquet file to save training history'},
    'plateau-reduce-factor': {
        'help': 'Factor multiplied by current learning rate ' 'when there is a plateau',
        'type': float,
        'default': 0.9,
    },
    'plateau-reduce-patience': {
        'help': 'Minimum non-improvements to reduce lr',
        'type': int,
        'default': 10,
    },
    'early-stop-patience': {
        'help': 'Minimum plateau reduces to stop training',
        'type': int,
        'default': 3,
    },
    'divergence-monitoring-steps': {
        'help': 'Steps after which, plateau and early stopping are active',
        'type': int,
    },
    'quiet': {
        'help': 'If specified, no messages by the agent will be displayed'
        '\nto the console',
        'action': 'store_true',
    },
}

train_args = {
    'target-reward': {
        'help': 'Target reward when reached, training is stopped',
        'type': int,
    },
    'max-steps': {
        'help': 'Maximum number of environment steps, when reached, training is stopped',
        'type': int,
    },
    'monitor-session': {'help': 'Wandb session name'},
}

play_args = {
    'video-dir': {'help': 'Path to directory to save the resulting gameplay video'},
    'render': {
        'help': 'If specified, the gameplay will be rendered',
        'action': 'store_true',
    },
    'frame-dir': {'help': 'Path to directory to save game frames'},
    'frame-delay': {
        'help': 'Delay between rendered frames',
        'type': float,
        'default': 0,
    },
    'action-idx': {
        'help': 'Index of action output by agent.model',
        'type': int,
        'default': 0,
    },
    'frame-frequency': {
        'help': 'If --frame-dir is specified, save frames every n frames.',
        'type': int,
        'default': 1,
    },
}

tune_args = {
    'trial-steps': {
        'help': 'Maximum steps for a trial',
        'type': int,
        'default': 500000,
    },
    'n-trials': {'help': 'Number of trials to run', 'type': int, 'default': 1},
    'study': {'help': 'Name of optuna study'},
    'storage': {'help': 'Database url'},
    'n-jobs': {'help': 'Parallel trials', 'type': int, 'default': 1},
    'warmup-trials': {
        'help': 'warmup trials before pruning starts',
        'type': int,
        'default': 5,
    },
    'non-silent': {
        'help': 'tensorflow, optuna and agent are silenced at trial start\n'
        'to avoid repetitive import messages at each trial start, unless\n'
        'this flag is specified',
        'action': 'store_true',
    },
}



import tensorflow as tf
from gym.spaces import Box
from tensorflow.keras.losses import MSE

from xagents.base import OffPolicy


class DDPG(OffPolicy):
    """
    Continuous control with deep reinforcement learning.
    https://arxiv.org/abs/1509.02971
    """

    def __init__(
        self,
        envs,
        actor_model,
        critic_model,
        buffers,
        gradient_steps=None,
        tau=0.05,
        step_noise_coef=0.1,
        **kwargs,
    ):
        """
        Initialize DDPG agent.
        Args:
            envs: A list of gym environments.
            actor_model: Actor separate model as a tf.keras.Model.
            critic_model: Critic separate model as a tf.keras.Model.
            buffers: A list of replay buffer objects whose length should match
                `envs`s'.
            gradient_steps: Number of iterations per train_step() call, if not
                specified, it defaults to the number of steps per most-recent
                finished episode per environment.
            tau: Tau constant used for syncing target model weights.
            policy_noise_coef: Coefficient multiplied by noise added to target actions.
            **kwargs: kwargs passed to super classes.
        """
        super(DDPG, self).__init__(envs, actor_model, buffers, **kwargs)
        self.assert_valid_env(envs[0], Box)
        self.actor = actor_model
        self.critic = critic_model
        self.policy_delay = 1
        self.gradient_steps = gradient_steps
        self.tau = tau
        self.step_noise_coef = step_noise_coef
        self.episode_steps = tf.Variable(tf.zeros(self.n_envs), False)
        self.step_increment = tf.ones(self.n_envs)
        self.output_models.append(self.critic)
        self.target_actor = tf.keras.models.clone_model(self.actor)
        self.target_critic = tf.keras.models.clone_model(self.critic)
        self.target_actor.set_weights(self.actor.get_weights())
        self.target_critic.set_weights(self.critic.get_weights())
        self.model_groups = [
            (self.actor, self.target_actor),
            (self.critic, self.target_critic),
        ]
        self.batch_dtypes = 5 * ['float32']

    def get_step_actions(self):
        """
        Get self.n_envs noisy actions to be stepped by self.envs

        Returns:
            actions
        """
        actions = self.actor(tf.numpy_function(self.get_states, [], tf.float64))
        actions += tf.random.normal(
            shape=(self.n_envs, self.n_actions), stddev=self.step_noise_coef
        )
        return tf.clip_by_value(actions, -1, 1)

    def sync_target_models(self):
        """
        Sync target actor, target critic 1, and target critic 2 weights
        with their respective models in self.model_groups.

        Returns:
            None
        """
        for model, target_model in self.model_groups:
            for var, target_var in zip(
                model.trainable_variables, target_model.trainable_variables
            ):
                target_var.assign((1 - self.tau) * target_var + self.tau * var)

    def update_actor_weights(self, states):
        """
        Update actor weights.
        Args:
            states: A tensor of shape (self.n_envs * total buffer batch size, *self.input_shape)

        Returns:
            None.
        """
        with tf.GradientTape() as tape:
            actor_loss = -tf.reduce_mean(
                self.critic(tf.concat([states, self.actor(states)], 1))
            )
        self.actor.optimizer.minimize(
            actor_loss, self.actor.trainable_variables, tape=tape
        )

    def update_critic_weights(self, states, actions, new_states, dones, rewards):
        """
        Update critic weights.
        Args:
            states: A tensor of shape (self.n_envs * total buffer batch size, *self.input_shape)
            actions: A tensor of shape (self.n_envs * total buffer batch size, self.n_actions)
            new_states: A tensor of shape (self.n_envs * total buffer batch size, *self.input_shape)
            dones: A tensor of shape (self.n_envs * total buffer batch size)
            rewards: A tensor of shape (self.n_envs * total buffer batch size)

        Returns:
            None
        """
        with tf.GradientTape() as tape:
            target_actions = self.target_actor(new_states)
            target_critic_input = tf.concat([new_states, target_actions], 1)
            target_value = self.target_critic(target_critic_input)
            target_value = rewards + tf.stop_gradient(
                (1 - dones) * self.gamma * target_value
            )
            critic_input = tf.concat([states, actions], 1)
            value = self.critic(critic_input)
            loss = MSE(value, target_value)
        self.critic.optimizer.minimize(loss, self.critic.trainable_variables, tape=tape)

    def update_weights(self, gradient_steps):
        """
        Run gradient steps and update both actor and critic weights according
            to self.policy delay for the given gradient steps.
        Args:
            gradient_steps: Number of iterations.

        Returns:
            None.
        """
        for gradient_step in range(int(gradient_steps)):
            states, actions, rewards, dones, new_states = tf.numpy_function(
                self.concat_buffer_samples, [], self.batch_dtypes
            )
            self.update_critic_weights(states, actions, new_states, dones, rewards)
            if gradient_step % self.policy_delay == 0:
                self.update_actor_weights(states)
                self.sync_target_models()

    @tf.function
    def train_step(self):
        """
        Perform 1 step which controls action_selection, interaction with environments
        in self.envs, batching and gradient updates.

        Returns:
            None
        """
        step_actions = self.get_step_actions()
        *_, dones, _ = tf.numpy_function(
            self.step_envs, [step_actions, True, True], self.batch_dtypes
        )
        for done_idx in tf.where(dones):
            gradient_steps = self.gradient_steps or self.episode_steps[done_idx[0]]
            self.update_weights(gradient_steps)
        self.episode_steps.assign(
            (self.episode_steps + self.step_increment) * (1 - dones)
        )



from xagents import a2c, ppo

a2c_args = a2c.cli_args
ppo_args = ppo.cli_args
trpo_args = {
    'actor-model': {'help': 'Path to actor model .cfg file'},
    'critic-model': {'help': 'Path to critic model .cfg file'},
    'max-kl': {
        'help': 'Maximum KL divergence used for calculating Lagrange multiplier',
        'type': float,
        'default': 1e-3,
        'hp_type': 'log_uniform',
    },
    'cg-iterations': {
        'help': 'Gradient conjugation iterations per train step',
        'type': int,
        'default': 10,
    },
    'cg-residual-tolerance': {
        'help': 'Gradient conjugation residual tolerance parameter',
        'type': float,
        'default': 1e-10,
        'hp_type': 'log_uniform',
    },
    'cg-damping': {
        'help': 'Gradient conjugation damping parameter',
        'type': float,
        'default': 1e-3,
        'hp_type': 'log_uniform',
    },
    'actor-iterations': {
        'help': 'Actor optimization iterations per train step',
        'type': int,
        'default': 10,
        'hp_type': 'int',
    },
    'critic-iterations': {
        'help': 'Critic optimization iterations per train step',
        'type': int,
        'default': 3,
        'hp_type': 'int',
    },
    'fvp-n-steps': {
        'help': 'Value used to skip every n-frames used to calculate FVP',
        'type': int,
        'default': 5,
        'hp_type': 'int',
    },
    'entropy-coef': {
        'help': 'Entropy coefficient for loss calculation',
        'type': float,
        'default': 0,
        'hp_type': 'log_uniform',
    },
    'lam': {
        'help': 'GAE-Lambda for advantage estimation',
        'type': float,
        'default': 1.0,
        'hp_type': 'log_uniform',
    },
    'n-steps': {
        'help': 'Transition steps',
        'type': int,
        'default': 512,
        'hp_type': 'categorical',
    },
}
cli_args = a2c_args.copy()
cli_args.update(ppo_args)
cli_args.update(trpo_args)
del cli_args['model']



import random
from collections import deque

import numpy as np


class BaseBuffer:
    """
    Base class for replay buffers.
    """

    def __init__(self, size, initial_size=None, batch_size=32):
        """
        Initialize replay buffer.
        Args:
            size: Buffer maximum size.
            initial_size: Buffer initial size to be filled before training starts.
                To be used by caller.
            batch_size: Size of the batch that should be used in get_sample() implementation.
        """
        assert (
            initial_size is None or initial_size > 0
        ), f'Buffer initial size should be > 0, got {initial_size}'
        assert size > 0, f'Buffer size should be > 0,  got {size}'
        assert batch_size > 0, f'Buffer batch size should be > 0, got {batch_size}'
        assert (
            batch_size <= size
        ), f'Buffer batch size `{batch_size}` should be <= size `{size}`'
        if initial_size:
            assert size >= initial_size, 'Buffer initial size exceeds max size'
        self.size = size
        self.initial_size = initial_size or size
        self.batch_size = batch_size
        self.current_size = 0

    def append(self, *args):
        """
        Add experience to buffer.
        Args:
            *args: Items to store, types are implementation specific.

        """
        raise NotImplementedError(
            f'append() should be implemented by {self.__class__.__name__} subclasses'
        )

    def get_sample(self):
        """
        Sample from stored experience.

        Returns:
            Sample as numpy array.
        """
        raise NotImplementedError(
            f'get_sample() should be implemented by {self.__class__.__name__} subclasses'
        )


class ReplayBuffer1(BaseBuffer):
    """
    deque-based replay buffer that holds state transitions
    """

    def __init__(self, size, **kwargs):
        """
        Initialize replay buffer.
        Args:
            size: Buffer maximum size.
            **kwargs: kwargs passed to BaseBuffer.
        """
        super(ReplayBuffer1, self).__init__(size, **kwargs)
        self.main_buffer = deque(maxlen=size)
        self.temp_buffer = []

    def append(self, *args):
        """
        Append experience and auto-allocate to temp buffer / main buffer(self)
        Args:
            *args: Items to store

        Returns:
            None
        """
        self.main_buffer.append(args)
        self.current_size = len(self.main_buffer)

    def get_sample(self):
        """
        Sample from stored experience.

        Returns:
            Same number of args passed to append, having self.batch_size as
            first shape.
        """
        memories = random.sample(self.main_buffer, self.batch_size)
        if self.batch_size > 1:
            return [np.array(item) for item in zip(*memories)]
        return memories[0]


class ReplayBuffer2(BaseBuffer):
    """
    numpy-based replay buffer added for compatibility with tensorflow shortcomings.
    """

    def __init__(self, size, slots, **kwargs):
        """
        Initialize replay buffer.

        Args:
            size: Buffer maximum size.
            slots: Number of args that will be passed to self.append()
            **kwargs: kwargs passed to BaseBuffer.
        """
        super(ReplayBuffer2, self).__init__(size, **kwargs)
        self.slots = [np.array([])] * slots
        self.current_size = 0

    def append(self, *args):
        """
        Add experience to buffer.
        Args:
            *args: Items to store.

        Returns:
            None
        """
        for i, arg in enumerate(args):
            if not isinstance(arg, np.ndarray):
                arg = np.array([arg])
            if not self.slots[i].shape[0]:
                self.slots[i] = np.zeros((self.size, *arg.shape), arg.dtype)
            self.slots[i][self.current_size % self.size] = arg.copy()
        if self.current_size < self.size:
            self.current_size += 1

    def get_sample(self):
        """
        Sample from stored experience.

        Returns:
            Same number of args passed to append, having self.batch_size as
            first shape.
        """
        indices = np.random.randint(
            0, min(self.current_size, self.size), self.batch_size
        )
        return [slot[indices] for slot in self.slots]



import argparse
import random
import string

import pytest
import tensorflow as tf
from gym.spaces import Discrete

import xagents
from xagents.tests.utils import assert_flags_displayed, get_expected_flags
from xagents.utils.common import create_model


@pytest.mark.usefixtures('executor', 'envs', 'envs2')
class TestExecutor:
    """
    Tests for command line options.
    """

    def test_display_section(self, section, capsys):
        """
        Check if appropriate flags are displayed in help menu.
        Args:
            section: Tuple of argument group name and respective dict of args.
            capsys: _pytest.capture.CaptureFixture
        """
        self.executor.display_section(*section)
        cap = capsys.readouterr().out
        assert_flags_displayed(cap, *section)

    @staticmethod
    def assert_base_displayed(cap):
        """
        Assert base help menu is displayed.
        Args:
            cap: Text displayed to the console.

        Returns:
            None
        """
        for keyword in [
            'xagents',
            'Usage',
            'xagents <command> <agent> [options] [args]',
            'Available commands:',
        ]:
            assert keyword in cap

    def test_display_commands(self, section, capsys):
        """
        Test basic and extended help display.
        Args:
            section: Tuple of argument group name and respective dict of args.
            capsys: _pytest.capture.CaptureFixture
        """
        self.executor.display_commands({section[0]: section[1]})
        cap = capsys.readouterr().out
        self.assert_base_displayed(cap)
        if section:
            assert_flags_displayed(cap, *section)

    @staticmethod
    def add_arg_value(_action, _type, _nargs, hp_type, test_args, values, flag, tuning):
        """
        Generate random values for the given arg according to its attributes.
        Args:
            _action: If True, True will be stored.
            _type: Argument type.
            _nargs: Argument number of expected values.
            hp_type: hyperparameter type.
            test_args: argv that is being created.
            values: A dictionary of flags - expected values
            flag: Flag to which a value is generated.
            tuning: If True, flags that have an `hp_type`
                will support multiple parameters which will
                be parsed according to their type.

        Returns:
            None
        """
        multiple_values = [random.randint(0, 100000) for _ in range(10)]
        multiple_values.sort()
        if not _action:
            if tuning and hp_type:
                if hp_type == 'categorical':
                    value = multiple_values
                else:
                    value = multiple_values[:2]
            elif _nargs:
                values_type = _type or str
                value = [values_type(item) for item in multiple_values]
            elif _type in [int, float]:
                value = random.randint(0, 100)
            else:
                value = ''.join(random.sample(string.ascii_lowercase, 10))
            if not isinstance(value, list):
                test_args.append(str(value))
                values[flag] = value
            else:
                for item in value:
                    test_args.append(f'{item}')
                values[flag] = value
        else:
            values[flag] = True

    @pytest.mark.parametrize('tuning', [True, False])
    def test_add_args(self, section, tuning):
        """
        Add given section to executor args, generate values, parse args
        and test if the values match.
        Args:
            section: Tuple of argument group name and respective dict of args.
        """
        parser = argparse.ArgumentParser()
        self.executor.add_args(section[1], parser, tuning)
        test_args = []
        values = {}
        for flag, options in section[1].items():
            _help = options.get('help')
            _default = options.get('default')
            _type = options.get('type')
            _action = options.get('action')
            _nargs = options.get('nargs')
            _hp_type = options.get('hp_type')
            test_args.append(f'--{flag}')
            self.add_arg_value(
                _action, _type, _nargs, _hp_type, test_args, values, flag, tuning
            )
        parsed_args = parser.parse_args(test_args)
        for attr, value in values.items():
            assert getattr(parsed_args, attr.replace('-', '_')) == value

    def test_maybe_create_agent_base_display(self, capsys):
        """
        Ensure only base help menu is displayed.
        Args:
            capsys: _pytest.capture.CaptureFixture
        """
        self.executor.maybe_create_agent([])
        cap = capsys.readouterr().out
        assert 'flag' not in cap
        self.assert_base_displayed(cap)

    def test_maybe_create_agent_invalid_command(self):
        """
        Ensure exception is raised for a valid command.
        """
        with pytest.raises(AssertionError, match=r'Invalid command'):
            self.executor.maybe_create_agent(['invalid'])

    def test_maybe_create_agent_command_agent(self, command, agent_id, capsys):
        """
        Ensure command only / command + agent flags help is displayed accordingly.
        Args:
            command: One of the commands available in xagents.commands
            agent_id: One of the agent ids available in xagents.agents
            capsys: _pytest.capture.CaptureFixture
        """
        test_args = [[command], [command, agent_id]]
        for argv in test_args:
            expected_flags = get_expected_flags(argv)
            self.executor.maybe_create_agent(argv)
            cap = capsys.readouterr().out
            for flag in expected_flags:
                assert f'--{flag}' in cap

    def test_maybe_create_agent_no_display(self, command, agent_id, capsys):
        """
        Ensure nothing is displayed if the help menu is not invoked using
        relevant args.
        Args:
            command: One of the commands available in xagents.commands
            agent_id: One of the agent ids available in xagents.agents
            capsys: _pytest.capture.CaptureFixture
        """
        argv = [command, agent_id, '--env', 'test-env']
        self.executor.maybe_create_agent(argv)
        assert not capsys.readouterr().out
        assert self.executor.agent_id == agent_id
        assert self.executor.command == command

    @pytest.mark.parametrize('tuning', [True, False])
    def test_parse_known_args(self, command, agent_id, tuning):
        """
        Ensure agent kwargs + non-agent kwargs + command kwargs match
        the expected ones given command and agent + minimum other flags.
        Args:
            command: One of the commands available in xagents.commands
            agent_id: One of the agent ids available in xagents.agents
            tuning: If True, flags that have an `hp_type`
                will support multiple parameters which will
                be parsed according to their type.
        """
        self.executor.command = command
        self.executor.agent_id = agent_id
        argv = [command, agent_id, '--env', 'test-env']
        if command == 'train':
            with pytest.raises(AssertionError, match=r'train requires'):
                self.executor.parse_known_args(argv, tuning)
            argv.extend(['--target-reward', '18'])
        agent_args, non_agent_args, command_args = self.executor.parse_known_args(
            argv, tuning
        )
        unknown_argv = argv + ['--unknown-flag', 'unknown-value']
        with pytest.warns(UserWarning, match=r'Got unknown'):
            self.executor.parse_known_args(unknown_argv, tuning)
        actual = {**vars(agent_args), **vars(non_agent_args), **vars(command_args)}
        assert set(get_expected_flags(argv, True)) == set(actual.keys())

    def test_create_models(self, agent_id):
        """
        Test creation of models and ensure resulting output units match the expected.
        Args:
            agent_id: One of the agent ids available in xagents.agents
        """
        envs = self.envs if agent_id not in ['td3', 'ddpg'] else self.envs2
        expected_units = [
            envs[0].action_space.n
            if isinstance(envs[0].action_space, Discrete)
            else envs[0].action_space.shape[0]
        ]
        if agent_id == 'acer':
            expected_units.append(expected_units[-1])
        elif agent_id != 'dqn':
            expected_units.append(1)
        actual_units = []
        for model_type in ['model', 'actor_model', 'critic_model']:
            if model_type in xagents.agents[agent_id]:
                model = create_model(envs[0], agent_id, model_type)
                total_units = 2 if model_type == 'model' and agent_id != 'dqn' else 1
                output_units = [
                    layer.units
                    for layer in model.layers
                    if isinstance(layer, tf.keras.layers.Dense)
                ][-total_units:]
                actual_units.extend(output_units)
        assert expected_units == actual_units

    @pytest.mark.parametrize(
        'train_args',
        [
            {
                'args': 'train a2c --env PongNoFrameskip-v4 --entropy-coef 5 '
                '--value-loss-coef 33 --grad-norm 7 --checkpoints xyz.tf '
                '--reward-buffer-size 150 --n-steps 100 --gamma 0.88 '
                '--display-precision 3 --seed 55 '
                '--log-frequency 28 --n-envs 25 --lr 0.555 --opt-epsilon 0.3 '
                '--beta1 15 --beta2 12 --max-steps 1',
                'agent': {
                    'entropy_coef': 5,
                    'value_loss_coef': 33,
                    'grad_norm': 7,
                    'checkpoints': ['xyz.tf'],
                    'n_steps': 100,
                    'gamma': 0.88,
                    'display_precision': 3,
                    'seed': 55,
                    'log_frequency': 28,
                    'n_envs': 25,
                },
                'non_agent': {
                    'env': 'PongNoFrameskip-v4',
                    'agent': xagents.A2C,
                    'lr': 0.555,
                    'opt_epsilon': 0.3,
                    'beta1': 15,
                    'beta2': 12,
                },
            }
        ],
    )
    def test_train(self, capsys, train_args):
        """
        Do 1 train step and test explicit parameters.
        Args:
            capsys: _pytest.capture.CaptureFixture
            train_args: A dictionary of agent, non-agent and command args
                to be run by executor.
        """
        self.executor.execute(train_args['args'].split())
        assert 'Maximum steps exceeded' in capsys.readouterr().out
        for attr, value in train_args['agent'].items():
            assert getattr(self.executor.agent, attr) == value
        assert isinstance(self.executor.agent, train_args['non_agent']['agent'])
        assert (
            self.executor.agent.envs[0].unwrapped.spec.id
            == train_args['non_agent']['env']
        )
        assert (
            self.executor.agent.model.optimizer.learning_rate
            == train_args['non_agent']['lr']
        )
        assert (
            self.executor.agent.model.optimizer.epsilon
            == train_args['non_agent']['opt_epsilon']
        )
        assert (
            self.executor.agent.model.optimizer.beta_1
            == train_args['non_agent']['beta1']
        )
        assert (
            self.executor.agent.model.optimizer.beta_2
            == train_args['non_agent']['beta2']
        )



import numpy as np
import pytest

from xagents.utils.buffers import BaseBuffer, ReplayBuffer2


def get_buffer_kwargs(size, buffer_type, initial_size, batch_size):
    """
    Construct buffer creation kwargs.
    Args:
        size: Buffer max size
        buffer_type: ReplayBuffer1 or ReplayBuffer2
        initial_size: Buffer initial pre-train fill size.
        batch_size: Buffer batch size when get_sample() is called.

    Returns:
        buffer kwargs as dict.
    """
    buffer_kwargs = {
        'size': size,
        'initial_size': initial_size,
        'batch_size': batch_size,
    }
    if buffer_type == ReplayBuffer2:
        buffer_kwargs['slots'] = 1
    return buffer_kwargs


def assert_sample_shapes_match(buffer, shapes):
    """
    Ensure buffer sample shapes match the expected ones.
    Args:
        buffer: ReplayBuffer1 or ReplayBuffer2 instance.
        shapes: 1 observation list of shapes.
    """
    sample = buffer.get_sample()
    size = sample[0].shape[0]
    for shape, subsample in zip(shapes, sample):
        assert np.squeeze(subsample).shape == (size, *shape)


class TestBaseBuffer:
    @pytest.mark.parametrize(
        'size, initial_size, batch_size, exception, exception_kw',
        [
            [100, 0, 32, AssertionError, 'Buffer initial size should be > 0, got'],
            [-100, 100, 32, AssertionError, 'Buffer size should be > 0'],
            [10, None, 32, AssertionError, 'should be <= size'],
            [100, 200, 32, AssertionError, 'Buffer initial size exceeds max size'],
            [100, 50, 32, None, None],
            [100, None, 32, None, None],
        ],
    )
    def test_sizes(
        self, buffer_type, size, initial_size, batch_size, exception, exception_kw
    ):
        """
        Test size parameters match the expected ones, and proper exceptions
        are raised.

        Args:
            buffer_type: ReplayBuffer1 or ReplayBuffer2
            size: Buffer max size.
            initial_size: Buffer initial pre-train fill size.
            batch_size: Buffer batch size when get_sample() is called.
            exception: Exception raised due to improper size parameters.
            exception_kw: exception keywords that are expected to be displayed.
        """
        buffer_kwargs = get_buffer_kwargs(size, buffer_type, initial_size, batch_size)
        if exception:
            with pytest.raises(exception, match=fr'{exception_kw}'):
                buffer_type(**buffer_kwargs)
        else:
            buffer = buffer_type(**buffer_kwargs)
            assert buffer.size == size
            assert (
                buffer.initial_size == buffer.size if not initial_size else initial_size
            )
            assert buffer.batch_size == batch_size

    def test_abstract_methods(self):
        """
        Test Ensure BaseBuffer abstract methods raise NotImplementedError
        """
        buffer = BaseBuffer(32)
        with pytest.raises(NotImplementedError, match=r'should be implemented'):
            buffer.append(1)
        with pytest.raises(NotImplementedError, match=r'should be implemented'):
            buffer.get_sample()


@pytest.mark.usefixtures('buffer1', 'observations')
class TestBuffer1:
    """
    Test ReplayBuffer1 methods.
    """

    @pytest.mark.parametrize(
        'args',
        [
            [1, 2],
            [1, 2, 3, 4],
            [np.random.randint(0, 1000, 10) for _ in range(5)],
            [np.random.randint(0, 1000, 10), 15, 11],
            [[np.random.randint(0, 1000, 10) for _ in range(5)]],
        ],
    )
    def test_append(self, args):
        """
        Test ReplayBuffer1.append
        Args:
            args: list of dummy observations.
        """
        self.buffer.append(*args)
        for arg, saved in zip(args, self.buffer.main_buffer[0]):
            if isinstance(arg, np.ndarray):
                assert (arg == saved).all()
            else:
                assert arg == saved

    def test_get_sample(self):
        """
        Test sample shapes.
        """
        shapes = [np.squeeze(item).shape for item in self.observations[0]]
        self.buffer.main_buffer.extend(self.observations)
        assert_sample_shapes_match(self.buffer, shapes)


@pytest.mark.usefixtures('observations', 'buffer2')
class TestBuffer2:
    """
    Test ReplayBuffer2 methods.
    """

    def test_append(self):
        """
        Ensure observations being stored properly.
        """
        expected_results = [[] for _ in range(5)]
        for observation in self.observations:
            self.buffer.append(*observation)
            for item, expected_items in zip(observation, expected_results):
                if not isinstance(item, np.ndarray):
                    expected_items.append([item])
                else:
                    expected_items.append(item)
        size = len(self.observations)
        for expected, actual in zip(expected_results, self.buffer.slots):
            assert (np.array(expected) == actual[:size]).all()

    def test_get_sample(self):
        """
        Test sample shapes.
        """
        shapes = [np.squeeze(item).shape for item in self.observations[0]]
        for observation in self.observations:
            self.buffer.append(*observation)
            assert_sample_shapes_match(self.buffer, shapes)



import os
from concurrent.futures import ProcessPoolExecutor, as_completed

import numpy as np
import optuna
import tensorflow as tf

import xagents
from xagents.utils.cli import agent_args, non_agent_args, off_policy_args
from xagents.utils.common import create_agent


class Objective:
    """
    Objective function wrapper class.
    """

    def __init__(
        self, agent_id, agent_known_args, non_agent_known_args, command_known_args
    ):
        """
        Initialize objective function parameters.
        Args:
            agent_id: One of the agent ids available in xagents.agents
            agent_known_args: argparse.Namespace, containing options passed to agent.
            non_agent_known_args: argparse.Namespace, containing options passed to
                environments, optimizer and other non-agent objects.
            command_known_args: argparse.Namespace, containing command-specific options.
        """
        self.agent_id = agent_id
        self.agent_args = agent_known_args
        self.non_agent_args = non_agent_known_args
        self.command_args = command_known_args
        self.arg_groups = [
            (
                vars(agent_known_args),
                {**agent_args, **xagents.agents[agent_id]['module'].cli_args},
                self.agent_args,
            ),
            (
                vars(non_agent_known_args),
                {**non_agent_args, **off_policy_args},
                self.non_agent_args,
            ),
        ]

    def set_trial_values(self, trial):
        """
        Set values of all available hyperparameters that are specified
        to be tuned.
        Args:
            trial: optuna.trial.Trial

        Returns:
            None
        """
        for parsed_args, default_args, namespace in self.arg_groups:
            for arg, possible_values in parsed_args.items():
                hp_type = default_args[arg.replace('_', '-')].get('hp_type')
                trial_value = possible_values
                if isinstance(possible_values, list):
                    if hp_type and len(possible_values) == 1:
                        trial_value = possible_values[0]
                    elif hp_type == 'categorical':
                        trial_value = trial.suggest_categorical(arg, possible_values)
                    elif hp_type == 'log_uniform':
                        trial_value = trial.suggest_loguniform(arg, *possible_values)
                    elif hp_type == 'int':
                        trial_value = trial.suggest_int(arg, *possible_values)
                setattr(namespace, arg, trial_value)

    def __call__(self, trial):
        """
        Objective function called by optuna.study.Study.optimize
        Args:
            trial: optuna.trial.Trial

        Returns:
            Trial reward.
        """
        self.set_trial_values(trial)
        agent = create_agent(
            self.agent_id, vars(self.agent_args), vars(self.non_agent_args), trial
        )
        agent.fit(max_steps=self.command_args.trial_steps)
        trial_reward = np.around(np.mean(agent.total_rewards or [0]), 2)
        return trial_reward


def run_trial(
    agent_id,
    agent_known_args,
    non_agent_known_args,
    command_known_args,
):
    """
    Run one trial.
    Args:
        agent_id: One of the agent ids available in xagents.agents
        agent_known_args: argparse.Namespace, containing options passed to agent.
        non_agent_known_args: argparse.Namespace, containing options passed to
            environments, optimizer and other non-agent objects.
        command_known_args: argparse.Namespace, containing command-specific options.

    Returns:
        None
    """
    if not command_known_args.non_silent:
        tf.get_logger().setLevel('ERROR')
        os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
        agent_known_args.quiet = True
    study = optuna.load_study(
        command_known_args.study, storage=command_known_args.storage
    )
    objective = Objective(
        agent_id, agent_known_args, non_agent_known_args, command_known_args
    )
    study.optimize(objective, n_trials=1)


def run_tuning(agent_id, agent_known_args, non_agent_known_args, command_known_args):
    """
    Run tuning session with tuning specs specified in command_known_args.
    Args:
        agent_id: One of the agent ids available in xagents.agents
        agent_known_args: argparse.Namespace, containing options passed to agent.
        non_agent_known_args: argparse.Namespace, containing options passed to
            environments, optimizer and other non-agent objects.
        command_known_args: argparse.Namespace, containing command-specific options.

    Returns:
        None
    """
    trial_kwargs = {
        'agent_id': agent_id,
        'agent_known_args': agent_known_args,
        'non_agent_known_args': non_agent_known_args,
        'command_known_args': command_known_args,
    }
    pruner = optuna.pruners.MedianPruner(command_known_args.warmup_trials)
    optuna.create_study(
        study_name=command_known_args.study,
        storage=command_known_args.storage,
        load_if_exists=True,
        direction='maximize',
        pruner=pruner,
    )
    for _ in range(command_known_args.n_trials // command_known_args.n_jobs):
        with ProcessPoolExecutor(command_known_args.n_jobs) as executor:
            future_trials = [
                executor.submit(run_trial, **trial_kwargs)
                for _ in range(command_known_args.n_jobs)
            ]
            for future_trial in as_completed(future_trials):
                future_trial.result()



from xagents import ddpg

ddpg_args = ddpg.cli_args
td3_args = {
    'policy-delay': {
        'help': 'Delay after which, actor weights and target models will be updated',
        'type': int,
        'default': 2,
        'hp_type': 'categorical',
    },
    'policy-noise-coef': {
        'help': 'Coefficient multiplied by noise added to target actions',
        'type': float,
        'default': 0.2,
        'hp_type': 'log_uniform',
    },
    'noise-clip': {
        'help': 'Target noise clipping value',
        'type': float,
        'default': 0.5,
        'hp_type': 'log_uniform',
    },
}
cli_args = ddpg_args.copy()
cli_args.update(td3_args)



from setuptools import find_packages, setup

install_requires = [
    'tensorflow_probability==0.13.0',
    'opencv_python_headless==4.5.3.56',
    'tensorflow_addons==0.14.0',
    'numpy==1.19.5',
    'pandas==1.3.2',
    'gym[atari,box2d]==0.19.0',
    'tensorflow==2.6.0',
    'wandb==0.12.1',
    'tabulate==0.8.9',
    'pyglet==1.5.15',
    'pytest==6.2.5',
    'pyarrow==5.0.0',
    'fastparquet==0.7.1',
    'matplotlib==3.4.3',
    'optuna==2.9.1',
    'termcolor==1.1.0',
]

setup(
    name='xagents',
    version='1.0.1',
    packages=find_packages(),
    url='https://github.com/schissmantics/xagents',
    license='MIT',
    author='schissmantics',
    author_email='schissmantics@outlook.com',
    description='Implementations of deep reinforcement learning algorithms in tensorflow 2.5',
    include_package_data=True,
    setup_requires=['numpy==1.19.5'],
    install_requires=install_requires,
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'xagents=xagents.cli:execute',
        ],
    },
)



import random
from time import perf_counter

import gym
import numpy as np
import optuna
import pandas as pd
import pytest
import tensorflow as tf
import wandb
from gym.spaces import Discrete
from optuna import Trial
from optuna.distributions import UniformDistribution

import xagents
from xagents import A2C, ACER, DDPG, DQN, PPO, TD3, TRPO
from xagents.base import BaseAgent, OffPolicy, OnPolicy
from xagents.utils.buffers import ReplayBuffer1
from xagents.utils.common import (create_buffers, create_models, get_wandb_key,
                                  write_from_dict)


class DummyAgent(BaseAgent):
    """
    Lightweight agent for testing time consuming tasks.
    """

    def __init__(self, envs, model, **kwargs):
        """
        Initialize dummy agent.
        Args:
            envs: A list of gym environments which are meant as placeholders.
            model: tf.keras.Model which serves as a placeholder.
            **kwargs: kwargs to be passed to super classes.
        """
        super(DummyAgent, self).__init__(envs, model, **kwargs)

    def at_step_start(self):
        """
        Execute steps that will run before self.train_step()
        which consist of printing a few words for testing purposes.

        Returns:
            None
        """
        print(f'step starting')

    def train_step(self):
        """
        A lightweight train step for test convenience purposes.

        Returns:
            None
        """
        self.steps += 1
        print(f'train_step')

    def at_step_end(self):
        """
        Execute steps that will run after self.train_step()
        which consist of printing a few words for testing purposes.

        Returns:
            None
        """
        print(f'step ending')


@pytest.mark.usefixtures('envs', 'envs2', 'model', 'buffers', 'study')
class TestBase:
    """
    Tests for base agents and their methods.
    """

    model_counts = {
        DDPG: 2,
        TD3: 3,
        TRPO: 2,
        A2C: 1,
        ACER: 1,
        DQN: 1,
        PPO: 1,
        BaseAgent: 1,
        OnPolicy: 1,
        OffPolicy: 1,
    }

    def get_agent_kwargs(
        self, agent, envs=None, model=None, buffers=None, critic_model=None
    ):
        """
        Construct agent required kwargs according to the given agent.
        Args:
            agent: OnPolicy/OffPolicy subclass.
            envs: A list of gym environments, if not specified, the default
                class envs will be returned.
            model: tf.keras.Model, if not specified, the default class model
                will be returned.
            buffers: A list of replay buffers, if not specified, the default
                class buffers will be returned.
            critic_model: tf.keras.Model, if not specified, the default class
                model will be returned.

        Returns:
            agent_kwargs.
        """
        valid_envs = self.envs if agent not in [TD3, DDPG] else self.envs2
        agent_kwargs = {'envs': envs if envs is not None else valid_envs}
        buffers = buffers or self.buffers
        if self.model_counts[agent] > 1:
            agent_kwargs['actor_model'] = model or tf.keras.models.clone_model(
                self.model
            )
            agent_kwargs['critic_model'] = critic_model or tf.keras.models.clone_model(
                self.model
            )
            agent_kwargs['actor_model'].compile('adam')
            agent_kwargs['critic_model'].compile('adam')
        else:
            agent_kwargs['model'] = model or self.model
        if agent == xagents.ACER:
            for buffer in buffers:
                buffer.batch_size = 1
            agent_kwargs['buffers'] = buffers
        if issubclass(agent, OffPolicy) or agent == OffPolicy:
            agent_kwargs['buffers'] = buffers
        return agent_kwargs

    def test_no_envs(self, agent):
        """
        Test no given environments.
        Args:
            agent: OnPolicy/OffPolicy subclass.
        """
        with pytest.raises(AssertionError, match=r'No environments given'):
            agent_kwargs = self.get_agent_kwargs(agent, [])
            agent(**agent_kwargs)

    def test_display_message(self, agent, capsys):
        agent = agent(**self.get_agent_kwargs(agent))
        message = 'displayed message'
        agent.display_message(message)
        assert message in capsys.readouterr().out
        agent.quiet = True
        agent.display_message(message)
        assert not capsys.readouterr().out

    def test_seeds(self, agent):
        """
        Test random seeds (random/numpy/tensorflow/gym)
        Args:
            agent: OnPolicy/OffPolicy subclass.
        """
        test_seed = 1
        agent = agent(**self.get_agent_kwargs(agent))
        results1, results2 = set(), set()
        test_range = 1000000
        for results in [results1, results2]:
            agent.set_seeds(test_seed)
            results.add(random.randint(0, test_range))
            results.add(int(np.random.randint(0, test_range, 1)))
            results.add(int(tf.random.uniform([1], maxval=test_range)))
            sample = agent.envs[0].action_space.sample()
            if isinstance(sample, np.ndarray):
                sample = tuple(sample)
            results.add(sample)
        assert results1 == results2

    @pytest.mark.parametrize(
        'env_ids',
        [
            ('RiverraidNoFrameskip-v4', 'MountainCarContinuous-v0'),
            ('Robotank-ramDeterministic-v0', 'Pendulum-v0'),
            ('BeamRiderDeterministic-v4', 'LunarLanderContinuous-v2'),
            ('PongNoFrameskip-v4', 'BipedalWalker-v3'),
            ('GravitarNoFrameskip-v0', 'BipedalWalkerHardcore-v3'),
            ('Freeway-ramDeterministic-v0', 'CarRacing-v0'),
        ],
    )
    def test_n_actions(self, env_ids, agent):
        """
        Test if agent initializes the correct n_actions attribute.
        Args:
            env_ids: discrete/box environment id pair.
            agent: OnPolicy/OffPolicy subclass.
        """
        env_id = env_ids[0] if agent not in [TD3, DDPG] else env_ids[1]
        env = gym.make(env_id)
        agent = agent(
            **self.get_agent_kwargs(agent, [env for _ in range(len(self.envs))])
        )
        if isinstance(env.action_space, Discrete):
            assert agent.n_actions == env.action_space.n
        else:
            assert agent.n_actions == env.action_space.shape[0]

    def test_unsupported_env(self, agent):
        """
        Test currently unsupported environments including multi-discrete.
        Args:
            agent: OnPolicy/OffPolicy subclass.
        """
        with pytest.raises(AssertionError, match=r'Expected one of'):
            agent(
                **self.get_agent_kwargs(
                    agent, [gym.make('RepeatCopy-v0') for _ in range(len(self.envs))]
                )
            )

    def test_init_from_checkpoint(self, agent, tmp_path):
        """
        Test initialization from history checkpoint, and ensure
        metrics are retrieved with the proper values.
        Args:
            agent: OnPolicy/OffPolicy subclass.
            tmp_path: pathlib.PosixPath
        """
        columns = [
            'time',
            'mean_reward',
            'best_reward',
            'step',
            'episode_reward',
        ]
        invalid_data = [{'a': [1], 'b': [2]}, {'a': [3], 'b': [4]}]
        invalid_checkpoint = tmp_path / 'invalid_checkpoint.parquet'
        for invalid_item in invalid_data:
            write_from_dict(invalid_item, invalid_checkpoint)
        agent = agent(
            **self.get_agent_kwargs(agent), history_checkpoint=invalid_checkpoint
        )
        with pytest.raises(AssertionError, match=r'Expected the following columns'):
            agent.init_from_checkpoint()
        history_size = 10
        valid_history = []
        agent.history_checkpoint = valid_checkpoint = (
            tmp_path / 'valid_checkpoint.parquet'
        )
        for size in range(history_size):
            valid_data = {column: [random.randint(0, 100)] for column in columns}
            write_from_dict(valid_data, valid_checkpoint)
            valid_history.append(
                {column: value[0] for (column, value) in valid_data.items()}
            )
        valid_history = pd.DataFrame(valid_history)
        agent.init_from_checkpoint()
        last_row = valid_history.loc[valid_history['time'].idxmax()]
        assert agent.mean_reward == last_row['mean_reward']
        assert agent.best_reward == valid_history['best_reward'].max()
        history_start_steps = last_row['step']
        history_start_time = last_row['time']
        times = agent.training_start_time, perf_counter() - history_start_time
        assert np.isclose(*times, rtol=0.1)
        assert agent.last_reset_step == agent.steps == history_start_steps
        assert np.mean(agent.total_rewards) == last_row['episode_reward']
        assert agent.games == valid_history.shape[0]

    def test_update_history(self, agent, tmp_path):
        """
        Test training history update that should happen after each
        finished episode. Several test episode values are checked in
        the respective file and validated against what's expected.
        Args:
            agent: OnPolicy/OffPolicy subclass.
            tmp_path: pathlib.PosixPath
        """
        history_checkpoint = tmp_path / 'test_checkpoint.parquet'
        agent = agent(
            **self.get_agent_kwargs(agent),
            history_checkpoint=history_checkpoint.as_posix(),
        )
        agent.training_start_time = perf_counter()
        mean_rewards = []
        best_rewards = []
        episode_rewards = []
        steps = []
        times = []
        for _ in range(5):
            mean_rewards.append(random.randint(1, 1000))
            best_rewards.append(random.randint(1, 1000))
            episode_rewards.append(random.randint(1, 1000))
            steps.append(random.randint(1, 1000))
            times.append(perf_counter() - agent.training_start_time)
            agent.mean_reward = mean_rewards[-1]
            agent.best_reward = best_rewards[-1]
            agent.steps = steps[-1]
            agent.update_history(episode_rewards[-1])
        assert history_checkpoint.exists()
        actual = (
            pd.read_parquet(history_checkpoint.as_posix())
            .sort_values('time')
            .drop('time', axis=1)
            .reset_index(drop=True)
        )
        expected = pd.DataFrame([mean_rewards, best_rewards, episode_rewards, steps]).T
        expected.columns = actual.columns
        assert (actual == expected).all().all()

    def test_checkpoint(self, agent, tmp_path, capsys):
        """
        Test display after keywords that are expected after a checkpoint.
        Also test for the presence of the expected checkpoint .tf files
        that are usually saved during training and ensure the number of models
        per agent in self.model_counts matches the one required by the agent,
        otherwise, an error will be raised by the agent.
        Args:
            agent: OnPolicy/OffPolicy subclass.
            tmp_path: pathlib.PosixPath
            capsys: _pytest.capture.CaptureFixture
        """
        checkpoints = {
            (tmp_path / f'test_weights{i}.tf').as_posix()
            for i in range(self.model_counts[agent])
        }
        expected_filenames = set()
        for checkpoint in checkpoints:
            expected_filenames.add(f'{checkpoint}.index')
            expected_filenames.add(f'{checkpoint}.data-00000-of-00001')
        expected_filenames.add((tmp_path / 'checkpoint').as_posix())
        agent = agent(**self.get_agent_kwargs(agent), checkpoints=checkpoints)
        agent.checkpoint()
        assert not capsys.readouterr().out
        agent.mean_reward = 100
        agent.checkpoint()
        assert 'Best reward updated' in capsys.readouterr().out
        assert agent.best_reward == 100
        resulting_files = {item.as_posix() for item in tmp_path.iterdir()}
        assert expected_filenames == resulting_files

    def test_wrong_checkpoints(self, agent):
        """
        Ensure an error is raised for the wrong number of checkpoints.
        Args:
            agent: OnPolicy/OffPolicy subclass.
        """
        agent = agent(
            **self.get_agent_kwargs(agent),
            checkpoints=(self.model_counts[agent] + 1) * ['wrong_ckpt.tf'],
        )
        with pytest.raises(AssertionError, match=r'given output models, got'):
            agent.fit(18)

    @staticmethod
    def assert_progress_displayed(displayed):
        """
        Assert text displayed to the console has the expected keywords.
        Args:
            displayed: str
        """
        for item in (
            'time',
            'steps',
            'games',
            'speed',
            'mean reward',
            'best reward',
        ):
            assert item in displayed

    def test_display_metrics(self, capsys, agent):
        """
        Check metric names are printed to the console.
        Args:
            capsys: _pytest.capture.CaptureFixture
            agent: OnPolicy/OffPolicy subclass.
        """
        agent = agent(**self.get_agent_kwargs(agent))
        agent.training_start_time = perf_counter()
        agent.frame_speed = agent.mean_reward = agent.best_reward = 0
        agent.display_metrics()
        displayed = capsys.readouterr().out
        self.assert_progress_displayed(displayed)

    @pytest.mark.parametrize(
        'steps, total_rewards, mean_reward, best_reward, plateau_count, '
        'expected_fps, expected_mean, learning_rate,'
        'interval, early_stop_count',
        [
            [1000, [100, 130, 150], 100, 99, 0, 500, 126.67, 0.01, 2, 0],
            [2000, [10, 15, 20], 50, 52, 3, 1000, 15, 0.01, 2, 1],
            [600000, [400, 300, 20], 100, 105, 1, 600, 240, 0.05, 1000, 1],
            [100000, [14, 22], 100, 105, 2, 20000, 18, 0.1, 5, 1],
            [1000, [1, 2, 3, 4, 5], 10, 9, 2, 1000, 3, 0.1, 1, 2],
        ],
    )
    def test_update_metrics(
        self,
        agent,
        steps,
        total_rewards,
        mean_reward,
        best_reward,
        plateau_count,
        expected_fps,
        expected_mean,
        learning_rate,
        interval,
        early_stop_count,
        capsys,
    ):
        """
        Ensure accuracy of the metrics displayed as the training
            progresses.
        Args:
            agent: OnPolicy/OffPolicy subclass.
            steps: Agent total environment steps so far.
            total_rewards: A list of finished episode rewards.
            mean_reward: Previous mean reward.
            best_reward: Previous best reward.
            plateau_count: Number of times agent reached plateau.
            expected_fps: Expected frame speed calculated given the other values.
            expected_mean: Expected calculated reward mean
            learning_rate: Start learning rate, before update.
            interval: Seconds back in time to set last reset time.
            early_stop_count: Number of times plateau reached its maximum
                patience and learning rate was reduced.
            capsys: _pytest.capture.CaptureFixture
        """
        agent = agent(**self.get_agent_kwargs(agent))
        agent.last_reset_time = perf_counter() - interval
        agent.steps = steps
        agent.total_rewards.extend(total_rewards)
        agent.mean_reward = mean_reward
        agent.best_reward = best_reward
        agent.plateau_count = plateau_count
        agent.early_stop_count = early_stop_count
        agent.divergence_monitoring_steps = 500000
        for model in agent.output_models:
            model.optimizer.learning_rate.assign(learning_rate)
        agent.update_metrics()
        cap = capsys.readouterr().out
        expected_learning_rate = learning_rate
        if plateau_count >= agent.plateau_reduce_patience:
            expected_learning_rate = agent.plateau_reduce_factor * learning_rate
        assert np.isclose(agent.frame_speed, expected_fps, rtol=3)
        assert agent.mean_reward == expected_mean
        assert np.isclose(agent.model.optimizer.learning_rate, expected_learning_rate)
        if mean_reward > best_reward:
            assert agent.early_stop_count == 0
            assert agent.plateau_count == 0
        if plateau_count >= agent.plateau_reduce_patience:
            assert agent.early_stop_count == early_stop_count + 1
        if learning_rate != expected_learning_rate:
            assert 'Learning rate reduced' in cap
            assert agent.plateau_count == 0
            assert agent.early_stop_count == early_stop_count + 1
        elif mean_reward < best_reward and steps >= agent.divergence_monitoring_steps:
            assert agent.plateau_count == plateau_count + 1

    def test_report_rewards(self, agent):
        """
        Test rewards are being reported as expected, and ensure
        unpromising branches are raising the appropriate exception.
        Args:
            agent: OnPolicy/OffPolicy subclass.
        """
        agent = agent(**self.get_agent_kwargs(agent))
        start_values = np.sort(np.random.randint(0, 100, 5))
        trial1 = optuna.trial.create_trial(
            params={'x': 2.0},
            distributions={'x': UniformDistribution(0, 10)},
            intermediate_values={i: start_values[i] for i in range(len(start_values))},
            value=55,
        )
        self.study.add_trial(trial1)
        trial2 = Trial(
            self.study, self.study._storage.create_new_trial(self.study._study_id)
        )
        agent.trial = trial2
        expected_rewards = start_values * 10
        for reward in expected_rewards:
            agent.total_rewards.append(reward)
            agent.report_rewards()
        reported_rewards = self.study.trials[-1].intermediate_values
        assert len(reported_rewards) == len(expected_rewards)
        for i in range(len(reported_rewards)):
            assert np.mean(expected_rewards[: i + 1]) == reported_rewards[i]
        trial3 = Trial(
            self.study, self.study._storage.create_new_trial(self.study._study_id)
        )
        agent.trial = trial3
        agent.total_rewards.clear()
        agent.reported_rewards = 0
        agent.total_rewards.append(-1)
        with pytest.raises(optuna.exceptions.TrialPruned):
            agent.report_rewards()

    def test_check_episodes(self, capsys, agent):
        """
        Ensure progress is displayed when expected and flags
            are updated afterwards.
        Args:
            capsys: _pytest.capture.CaptureFixture
            agent: OnPolicy/OffPolicy subclass.
        """
        agent = agent(**self.get_agent_kwargs(agent))
        agent.total_rewards.append(0)
        agent.training_start_time = perf_counter()
        agent.last_reset_time = perf_counter()
        agent.check_episodes()
        assert not capsys.readouterr().out
        agent.done_envs += len(self.envs)
        agent.check_episodes()
        self.assert_progress_displayed(capsys.readouterr().out)
        assert not agent.done_envs

    @pytest.mark.parametrize(
        'expected, mean_reward, steps, target_reward, max_steps, early_stop_count',
        [
            (False, 10, 0, 20, None, 0),
            (False, 10, 10, 11, 11, 0),
            (True, 100, 1000, 90, 2000, 0),
            (True, 200, 120, 500, 100, 0),
            (True, 1, 1, 1, 1, 0),
            (True, 100, 600000, 200, None, 5),
        ],
    )
    def test_training_done(
        self,
        capfd,
        expected,
        mean_reward,
        steps,
        target_reward,
        max_steps,
        early_stop_count,
        agent,
    ):
        """
        Test training status (done / not done) and ensure respective expected
            actions take place.
        Args:
            capfd: _pytest.capture.CaptureFixture
            expected: Expected training status where True indicates training
                should be done.
            mean_reward: Agent `mean_reward` attribute to set.
            steps: Agent `steps` attribute to set
            target_reward: Agent `target_reward` attribute to set.
            max_steps: Agent `max_steps` attribute to set
            early_stop_count: Number of times plateau reached its maximum
                patience and learning rate was reduced.
            agent: OnPolicy/OffPolicy subclass.
        """
        agent = agent(**self.get_agent_kwargs(agent))
        agent.mean_reward = mean_reward
        agent.steps = steps
        agent.target_reward = target_reward
        agent.max_steps = max_steps
        agent.early_stop_count = early_stop_count
        status = agent.training_done()
        messages = ['Reward achieved in', 'Maximum steps exceeded', 'Early stopping']
        if not expected:
            assert not status and not capfd.readouterr().out
        else:
            assert status
            cap = capfd.readouterr().out
            assert any([message in cap for message in messages])

    def test_concat_buffer_samples(self, off_policy_agent):
        """
        Test multiple buffer samples are concatenated properly.
        Args:
            off_policy_agent: OffPolicy subclass.
        """
        buffers = [ReplayBuffer1(10, batch_size=1) for _ in range(len(self.envs))]
        agent = off_policy_agent(
            **self.get_agent_kwargs(off_policy_agent, buffers=buffers)
        )
        with pytest.raises(ValueError, match=r'Sample larger than population'):
            agent.concat_buffer_samples()
        seen = []
        agent.batch_dtypes = ['float64', 'float16', 'float32', 'int16', 'int32']
        for i, buffer in enumerate(agent.buffers):
            observation = [[i], [i * 10], [i * 100], [i * 1000], [i * 10000]]
            seen.append(observation)
            buffer.append(*observation)
        expected = np.squeeze(np.array(seen)).T
        result = np.array(agent.concat_buffer_samples())
        assert [
            item.dtype for item in agent.concat_buffer_samples()
        ] == agent.batch_dtypes
        assert (result == expected).all()

    @pytest.mark.parametrize(
        'get_observation, store_in_buffers',
        [[False, False], [True, False], [False, True], [True, True]],
    )
    def test_step_envs(self, get_observation, store_in_buffers, agent):
        """
        Test observations are returned / stored in buffers when this is expected.
        Args:
            get_observation: arg passed to step_envs()
            store_in_buffers: arg passed to step_envs()
            agent: OnPolicy/OffPolicy subclass.
        """
        for buffer in self.buffers:
            buffer.main_buffer.clear()
            buffer.current_size = 0
        agent = agent(**self.get_agent_kwargs(agent))
        actions = [env.action_space.sample() for env in agent.envs]
        observations = agent.step_envs(actions, get_observation, store_in_buffers)
        if hasattr(agent, 'buffers'):
            if store_in_buffers:
                assert all([buffer.current_size > 0 for buffer in agent.buffers])
            else:
                assert all([buffer.current_size == 0 for buffer in agent.buffers])
        if get_observation:
            assert observations
            assert len(set([item.shape for item in observations[0]])) == 1
        else:
            assert not observations
        assert agent.steps == len(self.envs)

    @pytest.mark.skipif(not get_wandb_key(), reason='Wandb api key not available')
    @pytest.mark.parametrize(
        'target_reward, max_steps, monitor_session',
        [[18, None, None], [22, 55, None], [None, 100, 'test_session']],
    )
    def test_init_training(
        self, agent, capsys, target_reward, max_steps, monitor_session, tmpdir
    ):
        """
        Test training initialization and ensure training flags are set and
        a monitoring session is registered if one is specified.
        Args:
            agent: OnPolicy/OffPolicy subclass.
            capsys: _pytest.capture.CaptureFixture
            target_reward: Reward when reached, training should stop.
            max_steps: Number of agent steps, if reached, training should stop.
            monitor_session: wandb session name.
            tmpdir: py._path.local.LocalPath
        """
        checkpoints = [f'test-{i}.tf' for i in range(self.model_counts[agent])]
        agent = agent(**self.get_agent_kwargs(agent))
        agent.checkpoints = checkpoints
        with tmpdir.as_cwd():
            agent.init_training(target_reward, max_steps, monitor_session)
        assert agent.target_reward == target_reward
        assert agent.max_steps == max_steps
        assert agent.last_reset_time
        assert agent.training_start_time
        if monitor_session:
            assert wandb.run.name == monitor_session
            capsys.readouterr()

    def test_train_step(self, base_agent):
        """
        Ensure an exception is raised when train_step() abstract method is called.
        Args:
            base_agent: A base agent class.
        """
        agent = base_agent(**self.get_agent_kwargs(base_agent))
        with pytest.raises(NotImplementedError, match=r'should be implemented by'):
            agent.train_step()

    def test_get_model_outputs(self, base_agent):
        """
        Test single and multiple model outputs.
        Args:
            base_agent: A base agent class.
        """
        inputs = np.random.random((10, *self.envs[0].observation_space.shape))
        agent = base_agent(**self.get_agent_kwargs(base_agent))
        single_model_result = agent.get_model_outputs(inputs, self.model)
        multi_model_result = agent.get_model_outputs(inputs, [self.model, self.model])
        assert single_model_result.shape == multi_model_result[0].shape
        assert np.isclose(
            single_model_result.numpy(), multi_model_result[0].numpy()
        ).all()

    def test_play(
        self,
        agent,
        capsys,
        tmp_path,
        max_steps=10,
    ):
        """
        Play 1 game, and ensure resulting video / frames are saved.
        Args:
            agent: OnPolicy/OffPolicy subclass.
            capsys: _pytest.capture.CaptureFixture
            tmp_path: pathlib.PosixPath
            max_steps: Number of agent steps, if reached, training should stop.
        """
        agent_kwargs = {}
        agent_id = agent.__module__.split('.')[1]
        envs = self.envs if agent_id not in ['td3', 'ddpg'] else self.envs2
        agent_kwargs['envs'] = envs
        models = create_models(xagents.agents[agent_id], envs[0], agent_id)
        agent_kwargs.update(models)
        if issubclass(agent, OffPolicy) or agent == ACER:
            agent_kwargs['buffers'] = create_buffers(agent_id, 1000, 32, len(envs))
        agent = agent(**agent_kwargs)
        video_dir = tmp_path / agent_id / 'video'
        frame_dir = tmp_path / agent_id / 'frames'
        agent.play(video_dir, False, frame_dir, 0, max_steps)
        expected_frames = [frame_dir / f'{i:05d}.jpg' for i in range(max_steps)]
        for expected_frame in expected_frames:
            assert expected_frame.exists()
        assert len([*video_dir.rglob('openai*.mp4')])
        assert 'Maximum steps' in capsys.readouterr().out

    def test_fit(self, capsys):
        """
        Test agent.fit() and ensure the proper order of execution.
        Args:
            capsys: _pytest.capture.CaptureFixture
        """
        agent = DummyAgent(self.envs, self.model)
        agent.fit(max_steps=1)
        displayed = capsys.readouterr().out
        assert displayed.strip().split('\n') == [
            'step starting',
            'train_step',
            'step ending',
            'Maximum steps exceeded',
        ]

    @pytest.mark.parametrize('supported_agent', [DQN, TD3, DDPG])
    def test_fill_buffers(self, supported_agent, capsys):
        """
        Test buffers are filled up to the correct size, and
        filling progress was displayed.
        Args:
            supported_agent: An OffPolicy subclass.
            capsys: _pytest.capture.CaptureFixture
        """
        initial_size = 100
        max_size = 500
        n_buffers = len(self.envs)
        buffers = [
            ReplayBuffer1(500, initial_size=initial_size) for _ in range(n_buffers)
        ]
        agent = supported_agent(
            **self.get_agent_kwargs(supported_agent, buffers=buffers)
        )
        agent.fill_buffers()
        for i, buffer in enumerate(buffers):
            assert (
                buffer.current_size
                == buffer.initial_size
                == len(buffer.main_buffer)
                == initial_size
            )
            assert buffer.size == buffer.main_buffer.maxlen == max_size
        assert f'{n_buffers}/{n_buffers}' in capsys.readouterr().out



import numpy as np
import tensorflow as tf

from xagents import PPO


class TRPO(PPO):
    """
    Trust Region Policy Optimization.
    https://arxiv.org/abs/1502.05477
    """

    def __init__(
        self,
        envs,
        actor_model,
        critic_model,
        max_kl=1e-3,
        cg_iterations=10,
        cg_residual_tolerance=1e-10,
        cg_damping=1e-3,
        actor_iterations=10,
        critic_iterations=3,
        fvp_n_steps=5,
        **kwargs,
    ):
        """
        Initialize TRPO agent.
        Args:
            envs: A list of gym environments.
            actor_model: Actor separate model as a tf.keras.Model.
            critic_model: Critic separate model as a tf.keras.Model.
            max_kl: Maximum KL divergence used for calculating Lagrange multiplier
                and actor weight updates.
            cg_iterations: Gradient conjugation maximum number of iterations per train step.
            cg_residual_tolerance: Gradient conjugation residual tolerance parameter,
                which controls proceeding/stopping iterations.
            cg_damping: Gradient conjugation damping parameter used for calculation
                of Fisher vector product(FVP).
            actor_iterations: Maximum number of actor optimizations steps per train step.
            critic_iterations: Maximum number of critic optimizations steps per train step.
            fvp_n_steps: Value used to skip every n-frames used to calculate FVP
            **kwargs: kwargs passed to super classes.
        """
        super(TRPO, self).__init__(
            envs,
            actor_model,
            **kwargs,
        )
        self.output_models.append(critic_model)
        self.old_actor = tf.keras.models.clone_model(self.model)
        self.actor = self.model
        self.critic = critic_model
        self.cg_iterations = cg_iterations
        self.cg_residual_tolerance = cg_residual_tolerance
        self.cg_damping = cg_damping
        self.max_kl = max_kl
        self.critic_iterations = critic_iterations
        self.actor_iterations = actor_iterations
        self.fvp_n_steps = fvp_n_steps

    @staticmethod
    def flat_to_weights(flat, trainable_variables, in_place=False):
        """
        Convert a flat tensor which should be of the same size as flattened
        `trainable_variables` concatenated to a list of variables and assign
        the new values(optional) to previous `trainable_variables` values.
        Args:
            flat: Flat tensor that is compatible with the given trainable variables
                shapes.
            trainable_variables: A list of tf.Variable objects that should be
                compatible with the given flat tensor.
            in_place: If True, the new values will be assigned to given variables.

        Returns:
            A list of reshaped variables/ an empty list if `in_place`.
        """
        updated_trainable_variables = []
        start_idx = 0
        for trainable_variable in trainable_variables:
            shape = trainable_variable.shape
            flat_size = tf.math.reduce_prod(shape)
            updated_trainable_variable = tf.reshape(
                flat[start_idx : start_idx + flat_size], shape
            )
            if in_place:
                trainable_variable.assign(updated_trainable_variable)
            else:
                updated_trainable_variables.append(updated_trainable_variable)
            start_idx += flat_size
        return updated_trainable_variables

    @staticmethod
    def weights_to_flat(to_flatten, trainable_variables=None):
        """
        Flatten and concatenate a list of gradients/tf.Variable objects.
        Args:
            to_flatten: A list of tf.Variable objects.
            trainable_variables: If specified and the list of gradients
                has None values, these values will be replaced by zeros
                in the flat output.

        Returns:
            Flat tensor.
        """
        if not trainable_variables:
            to_concat = [tf.reshape(non_flat, [-1]) for non_flat in to_flatten]
        else:
            to_concat = [
                tf.reshape(
                    non_flat
                    if non_flat is not None
                    else tf.zeros_like(trainable_variable),
                    [-1],
                )
                for (non_flat, trainable_variable) in zip(
                    to_flatten, trainable_variables
                )
            ]
        return tf.concat(to_concat, 0)

    def calculate_fvp(self, flat_tangent, states):
        """
        Calculate Fisher vector product.
        Args:
            flat_tangent: Flattened gradients obtained by `self.weights_to_flat()`
            states: States tensor expected by the actor model.

        Returns:
            FVP as a flat tensor.
        """
        with tf.GradientTape() as tape2:
            with tf.GradientTape() as tape1:
                kl_divergence, *_ = self.calculate_kl_divergence(states)
            kl_grads = tape1.gradient(kl_divergence, self.actor.trainable_variables)
            tangents = self.flat_to_weights(
                flat_tangent, self.actor.trainable_variables
            )
            gvp = tf.add_n(
                [
                    tf.reduce_sum(grad * tangent)
                    for (grad, tangent) in zip(kl_grads, tangents)
                ]
            )
        hessians_products = tape2.gradient(gvp, self.actor.trainable_variables)
        return (
            self.weights_to_flat(hessians_products, self.actor.trainable_variables)
            + self.cg_damping * flat_tangent
        )

    def conjugate_gradients(self, flat_grads, states):
        """
        Get conjugated gradients.
        Args:
            flat_grads: Flat gradient tensor.
            states: States tensor expected by the actor model.

        Returns:
            Conjugated gradients as a flat tensor.
        """
        p = tf.identity(flat_grads)
        r = tf.identity(flat_grads)
        x = tf.zeros_like(flat_grads)
        r_dot_r = tf.tensordot(r, r, 1)
        iterations = 0
        while tf.less(iterations, self.cg_iterations) and tf.greater(
            r_dot_r, self.cg_residual_tolerance
        ):
            z = self.calculate_fvp(p, states)
            v = r_dot_r / tf.tensordot(p, z, 1)
            x += v * p
            r -= v * z
            new_r_dot_r = tf.tensordot(r, r, 1)
            mu = new_r_dot_r / r_dot_r
            p = r + mu * p
            r_dot_r = new_r_dot_r
            iterations += 1
        return x

    def calculate_kl_divergence(self, states):
        """
        Calculate probability distribution of both new and old actor models
        and calculate Kullback–Leibler divergence.
        Args:
            states: States tensor expected by the actor models.

        Returns:
            Mean KL divergence, old distribution and new distribution.
        """
        old_actor_output = self.get_model_outputs(
            states, [self.old_actor, self.critic]
        )[4]
        new_actor_output = self.get_model_outputs(states, self.output_models)[4]
        old_distribution = self.get_distribution(old_actor_output)
        new_distribution = self.get_distribution(new_actor_output)
        return (
            tf.reduce_mean(old_distribution.kl_divergence(new_distribution)),
            old_distribution,
            new_distribution,
        )

    def calculate_losses(self, states, actions, advantages):
        """
        Calculate surrogate loss and KL divergence.
        Args:
            states: States tensor expected by the actor models.
            actions: Tensor of actions taken.
            advantages: Tensor of calculated advantages.

        Returns:
            Surrogate loss and KL divergence.
        """
        (
            kl_divergence,
            old_distribution,
            new_distribution,
        ) = self.calculate_kl_divergence(states)
        entropy = tf.reduce_mean(new_distribution.entropy())
        entropy_loss = self.entropy_coef * entropy
        ratio = tf.exp(
            new_distribution.log_prob(actions) - old_distribution.log_prob(actions)
        )
        surrogate_gain = tf.reduce_mean(ratio * advantages)
        surrogate_loss = surrogate_gain + entropy_loss
        return surrogate_loss, kl_divergence

    def at_step_start(self):
        """
        Execute steps that will run before self.train_step() which updates
        old actor weights with the optimized weights.

        Returns:
            None
        """
        self.old_actor.set_weights(self.actor.get_weights())

    def update_actor_weights(
        self,
        flat_weights,
        full_step,
        surrogate_loss,
        states,
        actions,
        advantages,
    ):
        """
        Calculate losses, and run actor iterations until reaching `actor_iterations`
        specified earlier or reaching stop/ok conditions.
        Args:
            flat_weights: A tensor of flattened actor weights.
            full_step: FVP / Lagrange multiplier.
            surrogate_loss: The lower bound of the original objective - the expected
                cumulative return of the policy
            states: States tensor expected by the actor models.
            actions: Tensor of actions taken.
            advantages: Tensor of calculated advantages.

        Returns:
            None
        """
        learning_rate = 1.0
        for _ in range(self.actor_iterations):
            updated_weights = flat_weights + full_step * learning_rate
            self.flat_to_weights(updated_weights, self.actor.trainable_variables, True)
            losses = new_surrogate_loss, new_kl_divergence = self.calculate_losses(
                states, actions, advantages
            )
            improvement = new_surrogate_loss - surrogate_loss
            ok_conditions = [
                np.isfinite(losses).all(),
                new_kl_divergence <= self.max_kl * 1.5,
                improvement > 0,
            ]
            if all(ok_conditions):
                break
            learning_rate *= 0.5
        else:
            self.flat_to_weights(flat_weights, self.actor.trainable_variables, True)

    def update_critic_weights(self, states, returns):
        """
        Calculate value loss and apply gradients.
        Args:
            states: States tensor expected by the critic model.
            returns: The cumulative returns tensor.

        Returns:
            None
        """
        for _ in range(self.critic_iterations):
            for (states_mb, returns_mb) in self.get_mini_batches(states, returns):
                with tf.GradientTape() as tape:
                    values = self.get_model_outputs(states_mb, self.output_models)[2]
                    value_loss = tf.reduce_mean(tf.square(values - returns_mb))
                grads = tape.gradient(value_loss, self.critic.trainable_variables)
                self.critic.optimizer.apply_gradients(
                    zip(grads, self.critic.trainable_variables)
                )

    @tf.function
    def train_step(self):
        """
        Perform 1 step which controls action_selection, interaction with environments
        in self.envs, batching and gradient updates.

        Returns:
            None
        """
        states, actions, returns, values, _ = tf.numpy_function(
            self.get_batch, [], 5 * [tf.float32]
        )
        advantages = returns - values
        advantages = (advantages - tf.reduce_mean(advantages)) / tf.math.reduce_std(
            advantages
        )
        with tf.GradientTape() as tape:
            surrogate_loss, kl_divergence = self.calculate_losses(
                states, actions, advantages
            )
        flat_grads = self.weights_to_flat(
            tape.gradient(surrogate_loss, self.actor.trainable_variables),
            self.actor.trainable_variables,
        )
        step_direction = self.conjugate_gradients(
            flat_grads, states[:: self.fvp_n_steps]
        )
        shs = 0.5 * tf.tensordot(
            step_direction,
            self.calculate_fvp(step_direction, states[:: self.fvp_n_steps]),
            1,
        )
        lagrange_multiplier = tf.math.sqrt(shs / self.max_kl)
        full_step = step_direction / lagrange_multiplier
        pre_actor_weights = self.weights_to_flat(
            self.actor.trainable_variables,
        )
        tf.numpy_function(
            self.update_actor_weights,
            [
                pre_actor_weights,
                full_step,
                surrogate_loss,
                states,
                actions,
                advantages,
            ],
            [],
        )
        self.update_critic_weights(states, returns)
